! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: odfdump_mod
!
! Authors:  Masaaki Sakano - SSC/LUX
!          
! Description: A tasklet to dump the contents of the current ODF.
! 
! Comments: TODO:
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module odfdump_mod

  use types
  use param
  use oal
  use errorHandling

  use oal_aux	! For printODFProposal()

  implicit none 

  private 

  ! Command-line parameters:
  !   NONE (This task utilises SAS_ODF environment variable.)

  public odfdump

contains 

  !---------------------------------------------------------------------
  subroutine odfdump
  !---------------------------------------------------------------------

    implicit none

    type(ProposalInfoType) :: proposalInfo

    !-------------------------------------------------------------------
    ! Start

    call OAL_proposalInfo(proposalInfo)
    call printODFProposal(proposalInfo)

  end subroutine odfdump

  ! !---------------------------------------------------------------------
  ! subroutine readParameters
  ! !---------------------------------------------------------------------
  ! 
  !   implicit none
  ! 
  !   !-------------------------------------------------------------------
  !   ! Start
  ! 
  ! end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine printODFProposal_old(proposalInfo)	! Relocated in oal_aux
  !---------------------------------------------------------------------
  ! print the ODF information to STDOUT for users benefit or debugging.
  ! This subroutine is moved from rgssources_mod.f90 (Ver.6.1.1)

    implicit none 

    type(ProposalInfoType), intent(in) :: proposalInfo  ! typedef from oal

    !-------------------------------------------------------------------
    ! Start:

    ! from /packages/oal/doc/oalsources/oalexample.f90
    write(*, *) '=== ODF proposal information ==='
    write(*, *) 'GO name                    : ', trim(proposalInfo%GoName)
    write(*, *) 'GO institute               : ', trim(proposalInfo%GoInstitute)
    write(*, *) 'GO address                 : ', trim(proposalInfo%GoAddress)
    write(*, *) 'GO e-mail                  : ', trim(proposalInfo%GoEmail)
    write(*, *) 'AO                         : ', proposalInfo%AO
    write(*, *) 'science type               : ', proposalInfo%scienceType
    write(*, *) 'target name                : ', trim(proposalInfo%targetName)
    write(*, *) 'target RA                  : ', proposalInfo%targetRa
    write(*, *) 'target Dec                 : ', proposalInfo%targetDec
    write(*, *) 'duration of observation [s]: ', proposalInfo%obsDuration
    write(*, *) 'boresight RA               : ', proposalInfo%boresightRa
    write(*, *) 'boresight Dec              : ', proposalInfo%boresightDec
    write(*, *) 'position angle constr (lo) : ', proposalInfo%posAngleConstLo
    write(*, *) 'position angle constr (hi) : ', proposalInfo%posAngleConstHi
    write(*, *) 'position angle orig. ref.  : ', proposalInfo%posAngleOrigRef
    write(*, *) 'EMOS1 priority             : ', proposalInfo%emos1Prio
    write(*, *) 'EMOS2 priority             : ', proposalInfo%emos2Prio
    write(*, *) 'EPN priority               : ', proposalInfo%epnPrio
    write(*, *) 'RGS1 priority              : ', proposalInfo%rgs1Prio
    write(*, *) 'RGS2 priority              : ', proposalInfo%rgs2Prio
    write(*, *) 'OM priority                : ', proposalInfo%omPrio

  end subroutine printODFProposal_old

end module odfdump_mod

