! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
MODULE parse_list_mod

  USE errorhandling
!  USE get_free_io_unit_mod
use ssc_misc

  IMPLICIT NONE

  PUBLIC

CONTAINS

    SUBROUTINE parse_list(in_str, out_array, in_delim)

    CHARACTER(LEN=*), DIMENSION(:), POINTER :: out_array
    CHARACTER(LEN=*)                        :: in_str
    CHARACTER(LEN=1), OPTIONAL, INTENT(IN)  :: in_delim

    INTEGER           :: in_file, nstr, i, iold, istr, stat

    CHARACTER(LEN=1024) :: dumm


    IF( in_str(1:1)=="@" )THEN
       
       ! If the parameter starts with an @, it is a file to be read

!       CALL get_free_io_unit(in_file)
call getFreeIoUnit(in_file)
       OPEN(in_file, FILE=in_str(2:), IOSTAT=stat)
       IF( stat /= 0 )THEN
          WRITE(errstr, *) "Can't open file ", TRIM(in_str(2:)), &
               ", error ", stat
          CALL error(" ")
       END IF

       nstr = -1
       DO
          nstr = nstr + 1
          READ(in_file, *, IOSTAT=stat) dumm
          IF( stat<0 )THEN
             EXIT
          ELSE IF( stat>0 )THEN
             WRITE(errstr, *) "Can't read file ", TRIM(in_str(2:)), &
                  ", error ", stat
             CALL error(" ")
          END IF
       END DO
       REWIND(in_file)

       ALLOCATE(out_array(nstr))
       DO i= 1, nstr
          READ(in_file, *, IOSTAT=stat) out_array(i)
          out_array(i) = TRIM(out_array(i))
          IF( stat>0 )THEN
             WRITE(errstr, *) "Can't read file ", TRIM(in_str(2:)), &
                  ", error ", stat
             CALL error(" ")
          END IF
       END DO
       
       CLOSE(in_file)

    ELSE

       ! Parse the parameter string for the file names

       IF( PRESENT(in_delim) )THEN
          DO i = 1, LEN(in_str)
             IF( in_str(i:i)==in_delim ) in_str(i:i) = " "
          ENDDO
       ENDIF

       nstr = 0
       DO i = 2, LEN(in_str)
          IF( in_str(i:i)==" " .AND. in_str(i-1:i-1)/=" ") nstr=nstr+1
       END DO
       IF( in_str(i-1:i-1)/=" " ) nstr=nstr+1

       ! Step thru the string in reverse order, so the white space ends up at the end
       ! and can be trimmed off.

       iold = LEN(in_str)
       istr = 0
       ALLOCATE(out_array(nstr))

       DO i = LEN(in_str)-1, 1, -1
          IF( in_str(i:i)==" " .AND. in_str(i+1:i+1)/=" " )THEN
             istr = istr + 1
             out_array(nstr-istr+1) = TRIM(in_str(i+1:iold))
             iold = i
          END IF
       END DO

       IF( in_str(1:1)/=" " )THEN
          out_array(1) = in_str(1:iold)
       END IF

    END IF

  END SUBROUTINE parse_list
    
END MODULE parse_list_mod
