! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
module reallocate

  use types

  implicit none

  public realloc

  interface realloc
     module procedure reallocSingle, reallocDouble, reallocInt8, &
          reallocInt16, reallocInt32, reallocIntStr, reallocIntBool
  end interface

contains

  !---------------------------------------------------------------------
  subroutine reallocSingle(p, n) 
  !---------------------------------------------------------------------

    real(kind=single), pointer, dimension(:) :: p, pnew
    integer, intent(in) :: n
    integer :: nold, lb

    !-------------------------------------------------------------------
    ! Start

  if(.not. associated(p)) return

    lb = lbound(p,1)
    allocate( pnew(lb:lb+n-1))
    nold = min(size(p),n) + lb - 1
    pnew(:nold) = p(:nold)
    deallocate(p)
    p => pnew

  end subroutine reallocSingle

  !---------------------------------------------------------------------
  subroutine reallocDouble(p, n) 
  !---------------------------------------------------------------------

    real(kind=double), pointer, dimension(:) :: p, pnew
    integer, intent(in) :: n
    integer :: nold, lb

    !-------------------------------------------------------------------
    ! Start

  if(.not. associated(p)) return

    lb = lbound(p,1)
    allocate( pnew(lb:lb+n-1))
    nold = min(size(p),n) + lb - 1
    pnew(:nold) = p(:nold)
    deallocate(p)
    p => pnew

  end subroutine reallocDouble

  !---------------------------------------------------------------------
  subroutine reallocInt8(p, n) 
  !---------------------------------------------------------------------

    integer(kind=int8), pointer, dimension(:) :: p, pnew
    integer, intent(in) :: n
    integer :: nold, lb

    !-------------------------------------------------------------------
    ! Start

  if(.not. associated(p)) return

    lb = lbound(p,1)
    allocate( pnew(lb:lb+n-1))
    nold = min(size(p),n) + lb - 1
    pnew(:nold) = p(:nold)
    deallocate(p)
    p => pnew

  end subroutine reallocInt8

  !---------------------------------------------------------------------
  subroutine reallocInt16(p, n) 
  !---------------------------------------------------------------------

    integer(kind=int16), pointer, dimension(:) :: p, pnew
    integer, intent(in) :: n
    integer :: nold, lb

    !-------------------------------------------------------------------
    ! Start

  if(.not. associated(p)) return

    lb = lbound(p,1)
    allocate( pnew(lb:lb+n-1))
    nold = min(size(p),n) + lb - 1
    pnew(:nold) = p(:nold)
    deallocate(p)
    p => pnew

  end subroutine reallocInt16

  !---------------------------------------------------------------------
  subroutine reallocInt32(p, n) 
  !---------------------------------------------------------------------

    integer(kind=int32), pointer, dimension(:) :: p, pnew
    integer, intent(in) :: n
    integer :: nold, lb

    !-------------------------------------------------------------------
    ! Start

  if(.not. associated(p)) return

    lb = lbound(p,1)
    allocate( pnew(lb:lb+n-1))
    nold = min(size(p),n) + lb - 1
    pnew(:nold) = p(:nold)
    deallocate(p)
    p => pnew

  end subroutine reallocInt32

  !---------------------------------------------------------------------
  subroutine reallocIntBool(p, n) 
  !---------------------------------------------------------------------

    logical(kind=bool), pointer, dimension(:) :: p, pnew
    integer, intent(in) :: n
    integer :: nold, lb

    !-------------------------------------------------------------------
    ! Start

  if(.not. associated(p)) return

    lb = lbound(p,1)
    allocate( pnew(lb:lb+n-1))
    nold = min(size(p),n) + lb - 1
    pnew(:nold) = p(:nold)
    deallocate(p)
    p => pnew

  end subroutine reallocIntBool

  !---------------------------------------------------------------------
  subroutine reallocIntStr(p, n) 
  !---------------------------------------------------------------------

    character(LEN=*), pointer, dimension(:) :: p
    character(LEN=LEN(p)), pointer, dimension(:) :: pnew
    integer, intent(in) :: n
    integer :: nold, lb

    !-------------------------------------------------------------------
    ! Start

  if(.not. associated(p)) return

    lb = lbound(p,1)
    allocate( pnew(lb:lb+n-1))
    nold = min(size(p),n) + lb - 1
    pnew(:nold) = p(:nold)
    deallocate(p)
    p => pnew

  end subroutine reallocIntStr

end module reallocate
