! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
program testcoordinate
  use errorhandling
  use types
  use dal		! For INTEGER_NULL

  use coordinate

  !! Followings are in ssclib
  use constants		! for Pi etc
  use wcs_aux		! for type(WcsT)
  use test_utils	! for isNearlyEqual()

  implicit none

  call test_angleBetweenCelCoords
  call test_raDecToPixels

contains

  !---------------------------------------------------------------------
  subroutine test_angleBetweenCelCoords
  !---------------------------------------------------------------------
  ! Tests of angleBetweenCelCoords().
  ! Note: 

    implicit none

    character(25)  :: errHead
    real(single) :: vecAS(2), vecBS(2), angleS
    real(double) :: vecAD(2), vecBD(2), angleD

    integer, parameter :: precision = 6

    !-------------------------------------------------------------------
    ! Start:

    !----------------------------------------------------------
    ! Radian
    !----------------------------------------------------------
    errHead = "AngleBetweenCelCoords001a"
    print *,"=== "//errHead//": Radian (Pi)(0), Double"

    vecAD = (/Pi,    0.0d0/)	! Pi defined in coordinates
    vecBD = (/0.0d0, 0.0d0/)
    angleD = angleBetweenCelCoords(vecAD, vecBD)
    if (.not. isNearlyEqual(angleD, Pi, precision)) then
      print *,"angle=",angleD
      call error("err"//errHead, '')
    end if

    !----------------------------------------------------------
    errHead = "AngleBetweenCelCoords002a"
    print *,"=== "//errHead//": Radian (Pi)(0), Single"

    vecAS = real(vecAD, single)
    vecBS = real(vecBD, single)
    angleS = angleBetweenCelCoords(vecAS, vecBS)
    if (.not. isNearlyEqual(angleS, Pi, precision)) then
      print *,"angle=",angleS
      call error("err"//errHead, '')
    end if

    !----------------------------------------------------------
    errHead = "AngleBetweenCelCoords003a"
    print *,"=== "//errHead//": Radian (0)(Pi), Double"

    vecAD = (/0.0d0, 0.0d0/)	! Pi defined in coordinates
    vecBD = (/Pi,    0.0d0/)
    angleD = angleBetweenCelCoords(vecAD, vecBD)
    if (.not. isNearlyEqual(angleD, Pi, precision)) then
      call error("err"//errHead, '')
    end if

    !----------------------------------------------------------
    errHead = "AngleBetweenCelCoords011a"
    print *,"=== "//errHead//": Radian-Dec (Pi/2)(0), Double"

    vecAD = (/0.0d0, Pi/2/)	! Pi defined in coordinates
    vecBD = (/0.0d0, 0.0d0/)
    angleD = angleBetweenCelCoords(vecAD, vecBD)
    if (.not. isNearlyEqual(angleD, Pi/2, precision)) then
      print *,"angle=",angleD
      call error("err"//errHead, '')
    end if

    !----------------------------------------------------------
    errHead = "AngleBetweenCelCoords012a"
    print *,"=== "//errHead//": Radian-Dec (0)(Pi/2), Double"

    vecAD = (/0.0d0, 0.0d0/)	! Pi defined in coordinates
    vecBD = (/0.0d0, Pi/2/)
    angleD = angleBetweenCelCoords(vecAD, vecBD)
    if (.not. isNearlyEqual(angleD, Pi/2, precision)) then
      print *,"angle=",angleD
      call error("err"//errHead, '')
    end if

    !----------------------------------------------------------
    errHead = "AngleBetweenCelCoords013a"
    print *,"=== "//errHead//": Radian-Dec (0, Pi/2), Double"

    vecAS = real(vecAD, single)
    vecBS = real(vecBD, single)
    angleS = angleBetweenCelCoords(vecAS, vecBS)
    if (.not. isNearlyEqual(angleS, Pi/2, precision)) then
      print *,"angle=",angleS
      call error("err"//errHead, '')
    end if

    !----------------------------------------------------------
    errHead = "AngleBetweenCelCoords014a"
    print *,"=== "//errHead//": Radian-Dec (Pi/2)(-Pi/2), Double"

    vecAD = (/0.0d0,  Pi/2/)	! Pi defined in coordinates
    vecBD = (/0.0d0, -Pi/2/)
    angleD = angleBetweenCelCoords(vecAD, vecBD)
    if (.not. isNearlyEqual(angleD, Pi, precision)) then
      print *,"angle=",angleD
      call error("err"//errHead, '')
    end if


    !----------------------------------------------------------
    ! Deg
    !----------------------------------------------------------
    errHead = "AngleBetweenCelCoords101a"
    print *,"=== "//errHead//": Deg-RA (0)(180), Single"

    vecAS = (/0.0,   0.0/)
    vecBS = (/180.0, 0.0/)
    angleS = angleBetweenCelCoords(vecAS, vecBS, isRadian=.false.)
    if (.not. isNearlyEqual(angleS, 180, precision)) then
      print *,"angle=",angleS
      call error("err"//errHead, '')
    end if

    !----------------------------------------------------------
    errHead = "AngleBetweenCelCoords102a"
    print *,"=== "//errHead//": Deg-RA (0)(180), Double"

    vecAD = real(vecAS, double)
    vecBD = real(vecBS, double)
    angleD = angleBetweenCelCoords(vecAD, vecBD, isRadian=.false.)
    if (.not. isNearlyEqual(angleD, 180, precision)) then
      print *,"angle=",angleD
      call error("err"//errHead, '')
    end if

    !----------------------------------------------------------
    errHead = "AngleBetweenCelCoords103a"
    print *,"=== "//errHead//": Deg-RA (0)(180), Single"

    vecAS = (/0.0,   0.0/)
    vecBS = (/180.0, 0.0/)
    angleS = angleBetweenCelCoords(vecAS, vecBS, isRadian=.false.)
    if (.not. isNearlyEqual(angleS, 180, precision)) then
      print *,"angle=",angleS
      call error("err"//errHead, '')
    end if

    !----------------------------------------------------------
    errHead = "AngleBetweenCelCoords104a"
    print *,"=== "//errHead//": Deg-Dec (-90)(90), Single"

    vecAS = (/0.0, -90.0/)
    vecBS = (/0.0,  90.0/)
    angleS = angleBetweenCelCoords(vecAS, vecBS, isRadian=.false.)
    if (.not. isNearlyEqual(angleS, 180, precision)) then
      print *,"angle=",angleS
      call error("err"//errHead, '')
    end if

    !----------------------------------------------------------
    errHead = "AngleBetweenCelCoords105a"
    print *,"=== "//errHead//": Deg-RaDec (10,45)(190,45), Single"

    vecAS = (/ 10.0,  45.0/)
    vecBS = (/190.0,  45.0/)
    angleS = angleBetweenCelCoords(vecAS, vecBS, isRadian=.false.)
    if (.not. isNearlyEqual(angleS, 90, precision)) then
      print *,"angle=",angleS
      call error("err"//errHead, '')
    end if

    !----------------------------------------------------------
    errHead = "AngleBetweenCelCoords106a"
    print *,"=== "//errHead//": Deg-RaDec (350,-45)(170,-45), Single"

    vecAS = (/350.0, -45.0/)
    vecBS = (/170.0, -45.0/)
    angleS = angleBetweenCelCoords(vecAS, vecBS, isRadian=.false.)
    if (.not. isNearlyEqual(angleS, 90, precision)) then
      print *,"angle=",angleS
      call error("err"//errHead, '')
    end if

    !----------------------------------------------------------
    errHead = "AngleBetweenCelCoords107a"
    print *,"=== "//errHead//": Deg-RaDec (350,-45)(350,-45), Single"

    vecAS = (/350.0, -45.0/)
    vecBS = (/350.0, -45.0/)

    angleS = angleBetweenCelCoords(vecAS, vecBS, isRadian=.false.)

    if (abs(angleS) > 1e-06) then
      print *,"angle=",angleS
      call error("err"//errHead, '')
    end if

    !----------------------------------------------------------
    errHead = "AngleBetweenCelCoords108a"
    print *,"=== "//errHead//": Deg-RaDec (1,-89.96)(179,-89.99), Single"

    vecAS = (/  1.0, -89.96/)
    vecBS = (/179.0, -89.99/)
    angleS = angleBetweenCelCoords(vecAS, vecBS, isRadian=.false.)
    if (abs(angleS) > 0.1) then
      print *,"angle=",angleS
      call error("err"//errHead, '')
    end if

    !----------------------------------------------------------
    errHead = "AngleBetweenCelCoords109a"
    print *,"=== "//errHead//": Deg-RaDec (1,-89.96)(179,-89.99), Double"

    vecAD = real(vecAS, double)
    vecBD = real(vecBS, double)
    angleD = angleBetweenCelCoords(vecAD, vecBD, isRadian=.false.)
    if (abs(angleD) > 0.1) then
      print *,"angle=",angleD
      call error("err"//errHead, '')
    end if

    !----------------------------------------------------------
    errHead = "AngleBetweenCelCoords110a"
    print *,"=== "//errHead//": Deg-RaDec (1,90)(179,90), Single"

    vecAS = (/  1.0, 90.0/)
    vecBS = (/179.0, 90.0/)
    angleS = angleBetweenCelCoords(vecAS, vecBS, isRadian=.false.)
    if (abs(angleS) > 1e-20) then
      print *,"angle=",angleS
      call error("err"//errHead, '')
    end if

    !----------------------------------------------------------
    errHead = "AngleBetweenCelCoords111a"
    print *,"=== "//errHead//": Deg-RaDec (1,-90)(179,-90), Single"

    vecAS = (/  1.0, -90.0/)
    vecBS = (/179.0, -90.0/)
    angleS = angleBetweenCelCoords(vecAS, vecBS, isRadian=.false.)
    if (abs(angleS) > 1e-20) then
      print *,"angle=",angleS
      call error("err"//errHead, '')
    end if

    !----------------------------------------------------------
    errHead = "AngleBetweenCelCoords112a"
    print *,"=== "//errHead//": Deg-RaDec (1,90)(179,90), Double"

    vecAD = (/  1.0d0, 90.0d0/)
    vecBD = (/179.0d0, 90.0d0/)
    angleD = angleBetweenCelCoords(vecAD, vecBD, isRadian=.false.)
    if (abs(angleD) > 1e-20) then
      print *,"angle=",angleD
      call error("err"//errHead, '')
    end if

  end subroutine test_angleBetweenCelCoords


  !---------------------------------------------------------------------
  subroutine test_raDecToPixels
  !---------------------------------------------------------------------
  ! Tests of raDecToPixels() and pixelsToRaDec().
  ! Note: 

    implicit none

    character(25)  :: errHead
    real(single) :: xPixel, yPixel
    real(double) :: raRadian, decRadian
    type(WcsT)   :: wcs	! defined in ssclib/wcs_aux

    integer, parameter :: precision = 5

    !-------------------------------------------------------------------
    ! Start:

    call initializeWcsT(wcs)	! defined in wcs_aux

    wcs%x%ctype = 'RA---TAN'	! CTYPE1
    wcs%x%worldUnit = 'deg'	! CUUIT1
    wcs%x%refPixels = 100.5	! CRPIX1
    wcs%x%refWorld  = 60.0	! CRVAL1
    wcs%x%matrixOnAxis = -1.1111111111111111111d-03	! CDELT1 (4arcsec/pixel)

    wcs%y%ctype = 'DEC--TAN'	! CTYPE2
    wcs%y%worldUnit = 'deg'	! CUUIT2
    wcs%y%refPixels = 200.5	! CRPIX2
    wcs%y%refWorld  = 0.0	! CRVAL2
    wcs%y%matrixOnAxis = 1.1111111111111111111d-03	! CDELT2 (4arcsec/pixel)

    !----------------------------------------------------------
    ! raDecToPixels and pixelsToRaDec
    !----------------------------------------------------------
    errHead = "RaDecToPixels001a"
    print *,"=== "//trim(errHead)//": Centre"

    raRadian  = wcs%x%refWorld * DegToRad
    decRadian = wcs%y%refWorld * DegToRad
    call raDecToPixels(raRadian, decRadian, wcs, xPixel, yPixel)

    if (.not. isNearlyEqual(xPixel, wcs%x%refPixels, precision)  .or. &
        .not. isNearlyEqual(yPixel, wcs%y%refPixels, precision) ) then
      print *,"(xPixel,yPixel)=",xPixel,yPixel
      call error("err"//errHead, '')
    end if

    !----------------------------------------------------------
    errHead = "RaDecToPixels001b"
    print *,"=== "//trim(errHead)//": Centre (inverse)"

    call pixelsToRaDec(xPixel, yPixel, wcs, raRadian, decRadian)

    if (.not. isNearlyEqual(raRadian, wcs%x%refWorld * DegToRad, precision)  .or. &
        (abs(decRadian) > 0.001) ) then
      print *,"(ra,dec[deg])=", raRadian*RadToDeg, decRadian*RadToDeg
      call error("err"//errHead, '')
    end if

    !----------------------------------------------------------
    errHead = "RaDecToPixels002a"
    print *,"=== "//trim(errHead)//": Shift to RA (negative) at DEC=0 only"

    raRadian  = (wcs%x%refWorld - 0.1d0) * DegToRad	! 6arcmin=90pixels shift
    decRadian = wcs%y%refWorld * DegToRad
    call raDecToPixels(raRadian, decRadian, wcs, xPixel, yPixel)

    if (.not. isNearlyEqual(xPixel, wcs%x%refPixels+90, precision)  .or. &
        .not. isNearlyEqual(yPixel, wcs%y%refPixels, precision) ) then
      print *,"(xPixel,yPixel)=",xPixel,yPixel
      call error("err"//errHead, '')
    end if

    !----------------------------------------------------------
    errHead = "RaDecToPixels002b"
    print *,"=== "//trim(errHead)//": Inverse: Shift to RA (negative) at DEC=0 only"

    call pixelsToRaDec(xPixel, yPixel, wcs, raRadian, decRadian)

    if (.not. isNearlyEqual(raRadian, (wcs%x%refWorld-0.1d0)*DegToRad, precision)  .or. &
        (abs(decRadian) > 0.001) ) then
      print *,"(ra,dec[deg])=", raRadian*RadToDeg, decRadian*RadToDeg
      call error("err"//errHead, '')
    end if

    !----------------------------------------------------------
    errHead = "RaDecToPixels003a"
    print *,"=== "//trim(errHead)//": Inverse: Shift to DEC (negative) only"

    raRadian  = wcs%x%refWorld * DegToRad
    decRadian = (wcs%y%refWorld - 0.1d0) * DegToRad	! 6arcmin=90pixels shift
    call raDecToPixels(raRadian, decRadian, wcs, xPixel, yPixel)

    if (.not. isNearlyEqual(xPixel, wcs%x%refPixels, precision)  .or. &
        .not. isNearlyEqual(yPixel, wcs%y%refPixels-90, precision) ) then
      print *,"(xPixel,yPixel)=",xPixel,yPixel
      call error("err"//errHead, '')
    end if

    !----------------------------------------------------------
    errHead = "RaDecToPixels003b"
    print *,"=== "//trim(errHead)//": Inverse: Shift to DEC (negative) only"

    call pixelsToRaDec(xPixel, yPixel, wcs, raRadian, decRadian)

    if (.not. isNearlyEqual(raRadian, wcs%x%refWorld*DegToRad, precision)  .or. &
        .not. isNearlyEqual(decRadian, (wcs%y%refWorld-0.1d0)*DegToRad, precision) ) then
      print *,"(ra,dec[deg])=", raRadian*RadToDeg, decRadian*RadToDeg
      call error("err"//errHead, '')
    end if


    !----------------------------------------------------------
    errHead = "RaDecToPixels004a"
    print *,"=== "//trim(errHead)//": Shift to RA (positive) only at DEC=60 deg."

    wcs%y%refWorld  = 60.0	! Massive change in DEC -----

    raRadian  = (wcs%x%refWorld + 0.1d0) * DegToRad	! 6arcmin => 90/2=45pixels shift (1/2=cos(60deg))
    decRadian = wcs%y%refWorld * DegToRad
    call raDecToPixels(raRadian, decRadian, wcs, xPixel, yPixel)	! =>  (xPixel,yPixel)=  55.5000038   2.0053401E+02

    if (.not. isNearlyEqual(xPixel, wcs%x%refPixels-45, precision)  .or. &
        .not. isNearlyEqual(yPixel, wcs%y%refPixels, 4) ) then
        ! .not. isNearlyEqual(yPixel, wcs%y%refPixels, precision) ) then	! DEC changes a little bit, which is expected.
      print *,"(xPixel,yPixel)=",xPixel,yPixel
      call error("err"//errHead, '')
    end if

    !----------------------------------------------------------
    errHead = "RaDecToPixels004b"
    print *,"=== "//trim(errHead)//": Inverse: Shift to RA (positive) only at DEC=60 deg."

    call pixelsToRaDec(xPixel, yPixel, wcs, raRadian, decRadian)

    if (.not. isNearlyEqual(raRadian, (wcs%x%refWorld+0.1d0)*DegToRad, precision)  .or. &
        .not. isNearlyEqual(decRadian, wcs%y%refWorld*DegToRad, precision) ) then
      print *,"(ra,dec[deg])=", raRadian*RadToDeg, decRadian*RadToDeg
      call error("err"//errHead, '')
    end if


  end subroutine test_raDecToPixels


end program testcoordinate
