! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Program: testdssrangesaux
!
! Authors:  Masaaki Sakano - SSC/LUX
!          
! Description: Test dss_ranges_aux module.
! 
! Comments:
!
!-
program testdssrangesaux
  !use cal	! for convertPiToEeV (and CAL_setState)
  use dal
  use dss	! for RangeT and BoundT
  use errorhandling
  ! use test_utils	! for isNearlyEqual()
  !use dss_aux
  use dss_ranges_aux	! for copyRanges

  implicit none

  integer(int32), parameter :: maxRepM = 100000	! for (memory) stress test.
  !integer(int32), parameter :: maxRepM = 100	! for (memory) stress test.
  ! It takes ???sec ish with aral for maxRep = 100000 (in 2008).
  integer(int32) :: iM

  ! type(DataSetT) :: inSetM
  ! type(BlockT) :: inBlockM
  ! type(ArrayT) :: imageArrayM
  ! type(DssT) :: dssPointerM
  ! 
  ! type(RangeT), pointer :: piRangesM(:) => null()
  ! 
  ! character(2), parameter :: axisNameM = 'PI'
  ! logical :: thereAreConstraintsM, noCompObeysConstraintsM, isNewRangeChainM
  ! type(dssCompSubsetType), pointer :: compSubsetsM(:) => null()
  ! type(DScompT) :: dssCompM
  ! type(RangeT), pointer :: compRangesM(:) => null()
  ! !integer :: compNum
  ! integer :: numCompsM = 12, compNumM, numCompSubsetsM, compSubsetNumM

  print *, 'Test test_checkRangeOverlap()'
  call test_checkRangeOverlap

  print *, 'Test test_andRanges()'
  do iM=1,maxRepM
     call test_andRanges
  end do

  print *, 'Test test_andRangesVecVec()'
  do iM=1,maxRepM
     call test_andRangesVecVec
  end do

  ! call dssRelease(dssPointerM)
  ! call release(inSetM)

contains

  !---------------------------------------------------------------------
  subroutine test_checkRangeOverlap
  !---------------------------------------------------------------------
    
    implicit none

    type(RangeT) :: rangeA, rangeB
    integer :: status

    !! From dss_ranges_aux.f90
    ! UNDECIDED  = -1,&
    ! BISTOOLOW  =  1,&
    ! BISTOOHIGH =  2,&
    ! BOVERLAPSA =  3,&
    ! BCONTAINSA =  4,&
    ! BISWITHINA =  5,&
    ! BEQUALSA   =  6,&

    !--------------
    ! A,B UNDEFINED for lower limits
    !--------------

! A:	<------>
! B:	<------>
    rangeA = setRange(5.0, 10.0, UNDEFINED, UNDEFINED)
    rangeB = setRange(7.0, 10.0, UNDEFINED, UNDEFINED)
    status = checkRangeOverlap(rangeA, rangeB)
    if (status  .ne. BEQUALSA) then
      write(errstr, *) "Value = ", status
      call error("RangeOverlapAllUndefined")
    end if

! A:	<------>
! B:	<-----?
    rangeA = setRange(5.0, 10.0, UNDEFINED, UNDEFINED)
    rangeB = setRange(7.0, 10.0, UNDEFINED, INCLUSIVE)
    status = checkRangeOverlap(rangeA, rangeB)
    if (status  .ne. BISWITHINA) then
      write(errstr, *) "Value = ", status
      call error("RangeOverlapAllUndefinedButBUpper")
    end if

! A:	<-----?
! B:	<------>
    rangeA = setRange(5.0, 10.0, UNDEFINED, INCLUSIVE)
    rangeB = setRange(7.0, 10.0, UNDEFINED, UNDEFINED)
    status = checkRangeOverlap(rangeA, rangeB)
    if (status  .ne. BCONTAINSA) then
      write(errstr, *) "Value = ", status
      call error("RangeOverlapAllUndefinedButAUpper")
    end if

! A:	<-----I
! B:	<-----I
    rangeA = setRange(5.0, 10.0, UNDEFINED, INCLUSIVE)
    rangeB = setRange(7.0, 10.0, UNDEFINED, INCLUSIVE)
    status = checkRangeOverlap(rangeA, rangeB)
    if (status  .ne. BEQUALSA) then
      write(errstr, *) "Value = ", status
      call error("RangeOverlapLowUndefinedSame1")
    end if

! A:	<-----o
! B:	<-----o
    rangeA = setRange(5.0, 10.0, UNDEFINED, EXCLUSIVE)
    rangeB = setRange(7.0, 10.0, UNDEFINED, EXCLUSIVE)
    status = checkRangeOverlap(rangeA, rangeB)
    if (status  .ne. BEQUALSA) then
      write(errstr, *) "Value = ", status
      call error("RangeOverlapLowUndefinedSame2")
    end if

! A:	<-----I
! B:	<-----o
    rangeA = setRange(5.0, 10.0, UNDEFINED, INCLUSIVE)
    rangeB = setRange(7.0, 10.0, UNDEFINED, EXCLUSIVE)
    status = checkRangeOverlap(rangeA, rangeB)
    if (status  .ne. BISWITHINA) then
      write(errstr, *) "Value = ", status
      call error("RangeOverlapLowUndefinedBExclusive")
    end if

! A:	<-----o
! B:	<-----I
    rangeA = setRange(5.0, 10.0, UNDEFINED, EXCLUSIVE)
    rangeB = setRange(7.0, 10.0, UNDEFINED, INCLUSIVE)
    status = checkRangeOverlap(rangeA, rangeB)
    if (status  .ne. BCONTAINSA) then
      write(errstr, *) "Value = ", status
      call error("RangeOverlapLowUndefinedAExclusive")
    end if

! A:	<------I
! B:	<-----I
    rangeA = setRange(5.0, 19.0, UNDEFINED, INCLUSIVE)
    rangeB = setRange(7.0, 10.0, UNDEFINED, INCLUSIVE)
    status = checkRangeOverlap(rangeA, rangeB)
    if (status  .ne. BISWITHINA) then
      write(errstr, *) "Value = ", status
      call error("RangeOverlapLowUndefinedALarger")
    end if

! A:	<-----I
! B:	<------I
    rangeA = setRange(5.0, 10.0, UNDEFINED, INCLUSIVE)
    rangeB = setRange(7.0, 19.0, UNDEFINED, INCLUSIVE)
    status = checkRangeOverlap(rangeA, rangeB)
    if (status  .ne. BCONTAINSA) then
      write(errstr, *) "Value = ", status
      call error("RangeOverlapLowUndefinedBLarger")
    end if

    !--------------
    ! A,B UNDEFINED for upper limits
    !--------------
! A:	<------>
! B:	?------>

! A:	?------>
! B:	<------>

! A:	I------>
! B:	I------>

! A:	o------>
! B:	o------>

! A:	 o----->
! B:	o------>

! A:	o------>
! B:	 o----->

    !--------------
    ! Just two paramters in the opposite sides are UNDEFINED
    !--------------
! A:	<------>
! B:	 I----I

! A:	 I----I
! B:	<------>

! A:	<-----I
! B:	 I----->

! A:	 I----->
! B:	<-----I

! A:	<-----I
! B:	       I----->

! A:	<-----I
! B:	      I----->

! A:	<-----o
! B:	      o----->

! A:	       I----->
! B:	<-----I

! A:	      I----->
! B:	<-----I

! A:	      o----->
! B:	<-----o


    !--------------
    ! No more than one paramter is UNDEFINED
    !--------------
! A:	<------I
! B:	 I-----I

! A:	<------o
! B:	 I-----o

! A:	<------o
! B:	 I-----I

! A:	<-----I
! B:	 I-----I

! A:	<-----I
! B:	       I-----I

! A:	<-----I
! B:	      o-----I

! A:	<-----o
! B:	      I-----I

! A:	<-----I
! B:	      I-----I

! A:	<-----o
! B:	      o-----I

! A:	I------>
! B:	I-----I

! A:	o------>
! B:	o-----I

! A:	o------>
! B:	I-----I

! A:	 I----->
! B:	I-----I

! A:	       I----->
! B:	I-----I

! A:	 I-----I
! B:	<------I

! A:	 I-----o
! B:	<------o

! A:	 I-----I
! B:	<------o

! A:	 I-----I
! B:	<-----I

! A:	       I-----I
! B:	<-----I

! A:	I-----I
! B:	I------>

! A:	o-----I
! B:	o------>

! A:	I-----I
! B:	o------>

! A:	I-----I
! B:	 I----->

! A:	I-----I
! B:	       I----->

    !--------------
    ! Everything is defined
    !--------------

! A:	I-----I
! B:	I-----I

! A:	o-----o
! B:	o-----o

! A:	I-----I
! B:	I-----o

! A:	I-----I
! B:	o-----I

! A:	o-----I
! B:	I-----I

! A:	I-----o
! B:	I-----I

! A:	I-----I
! B:	o-----o

! A:	o-----o
! B:	I-----I

! A:	o-----I
! B:	I-----o

! A:	I-----o
! B:	o-----I

! A:	I-----o
! B:	I-----I

! A:	I-----I
! B:	o-----I

! A:	I-----I
! B:	 I---I
    rangeA = setRange(1.0, 19.0, INCLUSIVE, INCLUSIVE)
    rangeB = setRange(3.0, 10.0, INCLUSIVE, INCLUSIVE)
    status = checkRangeOverlap(rangeA, rangeB)
    if (status  .ne. BISWITHINA) then
      write(errstr, *) "Value = ", status
      call error("RangeOverlapAllDefinedABigger")
    end if

! A:	 I---I
! B:	I-----I
    rangeA = setRange(3.0, 10.0, INCLUSIVE, INCLUSIVE)
    rangeB = setRange(1.0, 19.0, INCLUSIVE, INCLUSIVE)
    status = checkRangeOverlap(rangeA, rangeB)
    if (status  .ne. BCONTAINSA) then
      write(errstr, *) "Value = ", status
      call error("RangeOverlapAllDefinedBBigger")
    end if

! A:	I----I
! B:	 I----I
    rangeA = setRange(1.0, 10.0, INCLUSIVE, INCLUSIVE)
    rangeB = setRange(3.0, 19.0, INCLUSIVE, INCLUSIVE)
    status = checkRangeOverlap(rangeA, rangeB)
    if (status  .ne. BOVERLAPSA) then
      write(errstr, *) "Value = ", status
      call error("RangeOverlapAllDefinedBLargerOLap")
    end if

! A:	 I----I
! B:	I----I
    rangeA = setRange(3.0, 19.0, INCLUSIVE, INCLUSIVE)
    rangeB = setRange(1.0, 10.0, INCLUSIVE, INCLUSIVE)
    status = checkRangeOverlap(rangeA, rangeB)
    if (status  .ne. BOVERLAPSA) then
      write(errstr, *) "Value = ", status
      call error("RangeOverlapAllDefinedALargerOLap")
    end if

! A:	I----I
! B:	      I----I
    rangeA = setRange(1.0,  2.0, INCLUSIVE, INCLUSIVE)
    rangeB = setRange(9.0, 19.0, INCLUSIVE, INCLUSIVE)
    status = checkRangeOverlap(rangeA, rangeB)
    if (status  .ne. BISTOOHIGH) then
      write(errstr, *) "Value = ", status
      call error("RangeOverlapAllDefinedBLarger")
    end if

! A:	      I----I
! B:	I----I
    rangeA = setRange(9.0, 19.0, INCLUSIVE, INCLUSIVE)
    rangeB = setRange(1.0,  2.0, INCLUSIVE, INCLUSIVE)
    status = checkRangeOverlap(rangeA, rangeB)
    if (status  .ne. BISTOOLOW) then
      write(errstr, *) "Value = ", status
      call error("RangeOverlapAllDefinedALarger")
    end if

! A:	      o----I
! B:	I----o
    rangeA = setRange(9.0, 19.0, EXCLUSIVE, INCLUSIVE)
    rangeB = setRange(1.0,  2.0, INCLUSIVE, EXCLUSIVE)
    status = checkRangeOverlap(rangeA, rangeB)
    if (status  .ne. BISTOOLOW) then
      write(errstr, *) "Value = ", status
      call error("RangeOverlapAllDefinedAAbsLarger1")
    end if

! A:	      o----o
! B:	o----o

! A:	     I----o
! B:	o----o
    rangeA = setRange(9.0, 19.0, INCLUSIVE, EXCLUSIVE)
    rangeB = setRange(1.0,  9.0, EXCLUSIVE, EXCLUSIVE)
    status = checkRangeOverlap(rangeA, rangeB)
    if (status  .ne. BISTOOLOW) then
      write(errstr, *) "Value = ", status
      call error("RangeOverlapAllDefinedAJustLarger1")
    end if

! A:	     o----o
! B:	o----I
    rangeA = setRange(9.0, 19.0, EXCLUSIVE, EXCLUSIVE)
    rangeB = setRange(1.0,  9.0, EXCLUSIVE, INCLUSIVE)
    status = checkRangeOverlap(rangeA, rangeB)
    if (status  .ne. BISTOOLOW) then
      write(errstr, *) "Value = ", status
      call error("RangeOverlapAllDefinedAJustLarger2")
    end if

! A:	     o----o
! B:	o----o
    rangeA = setRange(9.0, 19.0, EXCLUSIVE, EXCLUSIVE)
    rangeB = setRange(1.0,  9.0, EXCLUSIVE, EXCLUSIVE)
    status = checkRangeOverlap(rangeA, rangeB)
    if (status  .ne. BISTOOLOW) then
      write(errstr, *) "Value = ", status
      call error("RangeOverlapAllDefinedAJustLarger3")
    end if

! A:	     I----o
! B:	o----I
    rangeA = setRange(3.0, 19.0, INCLUSIVE, EXCLUSIVE)
    rangeB = setRange(1.0,  3.0, EXCLUSIVE, INCLUSIVE)
    status = checkRangeOverlap(rangeA, rangeB)
    if (status  .ne. BOVERLAPSA) then
      write(errstr, *) "Value = ", status
      call error("RangeOverlapAllDefinedALargerJustOLap1")
    end if

! A:	o----o
! B:	     I----o

! A:	o----I
! B:	     o----o

! A:	o----o
! B:	     o----o

! A:	o----I
! B:	     I----o
    rangeA = setRange(1.0,  3.0, EXCLUSIVE, INCLUSIVE)
    rangeB = setRange(3.0, 19.0, INCLUSIVE, EXCLUSIVE)
    status = checkRangeOverlap(rangeA, rangeB)
    if (status  .ne. BOVERLAPSA) then
      write(errstr, *) "Value = ", status
      call error("RangeOverlapAllDefinedBLargerJustOLap1")
    end if

  end subroutine test_checkRangeOverlap


  !---------------------------------------------------------------------
  subroutine test_andRanges
  !---------------------------------------------------------------------
    
    implicit none

    type(RangeT) :: rangeA, rangeB, rangeR
    type(RangeT),  pointer              :: andedRanges(:)

    logical(bool), parameter :: doChecks = .false.


! A:	I----o
! B:	  o--I
    rangeA = setRange(5.0, 10.0, INCLUSIVE, EXCLUSIVE)
    rangeB = setRange(7.0, 10.0, EXCLUSIVE, INCLUSIVE)

    rangeR = setRange(7.0, 10.0, EXCLUSIVE, EXCLUSIVE)

    call andRanges(rangeA, rangeB, andedRanges, doChecks)	! call andRangesScalScalVec(rangeA, rangeB, andedRanges, doChecks)

    if (size(andedRanges) .ne. 1) then
      write(errstr, *) "Inconsistent andedRangesSize = ", size(andedRanges)
      call error("InconsistentAndedRangesSize1")
    else if (.not. areRangesEqual(andedRanges(1), rangeR)) then
      call dumpRanges(andedRanges(1))
      ! write(errstr, *) "Inconsistent andedRanges = " 
      call error("InconsistentAndedRanges1")
    end if

! A:	I----o
! B:	  o----I
    rangeA = setRange(5.0, 10.0, INCLUSIVE, EXCLUSIVE)
    rangeB = setRange(7.0, 15.0, EXCLUSIVE, INCLUSIVE)

    rangeR = setRange(7.0, 10.0, EXCLUSIVE, EXCLUSIVE)
    call andRanges(rangeA, rangeB, andedRanges, doChecks)	! call andRangesScalScalVec(rangeA, rangeB, andedRanges, doChecks)

    if (size(andedRanges) .ne. 1) then
      write(errstr, *) "Inconsistent andedRangesSize = ", size(andedRanges)
      call error("InconsistentAndedRangesSize2")
    else if (.not. areRangesEqual(andedRanges(1), rangeR)) then
      call dumpRanges(andedRanges(1))
      ! write(errstr, *) "Inconsistent andedRanges = " 
      call error("InconsistentAndedRanges2")
    end if

! A:	I----o
! B:	      o----I
    rangeA = setRange(5.0,  10.0, INCLUSIVE, EXCLUSIVE)
    rangeB = setRange(12.0, 15.0, EXCLUSIVE, INCLUSIVE)

    !rangeR = setRange(7.0, 10.0, EXCLUSIVE, EXCLUSIVE)
    call andRanges(rangeA, rangeB, andedRanges, doChecks)	! call andRangesScalScalVec(rangeA, rangeB, andedRanges, doChecks)

    if (size(andedRanges) .ne. 0) then
      write(errstr, *) "Inconsistent andedRangesSize = ", size(andedRanges)
      call error("InconsistentAndedRangesSize3")
    end if

    deallocate(andedRanges)

  end subroutine test_andRanges


  !---------------------------------------------------------------------
  subroutine test_andRangesVecVec
  !---------------------------------------------------------------------
    
    implicit none

    type(RangeT), dimension(2) :: rangeA, rangeB, rangeR
    type(RangeT),  pointer              :: andedRanges(:)

    logical(bool), parameter :: doChecks = .false.

    rangeA(1) = setRange(2.0, 4.0, INCLUSIVE, EXCLUSIVE)
    rangeA(2) = setRange(7.0, 9.0, EXCLUSIVE, INCLUSIVE)
    rangeB(1) = setRange(2.0, 8.0, INCLUSIVE, INCLUSIVE)
    rangeB(2) = setRange(15.0, 19.0, INCLUSIVE, INCLUSIVE)

    rangeR(1) = setRange(2.0, 4.0, INCLUSIVE, EXCLUSIVE)
    rangeR(2) = setRange(7.0, 8.0, EXCLUSIVE, INCLUSIVE)

    call andRanges(rangeA, rangeB, andedRanges, doChecks)	! call andRangesVecVecVec(rangeA, rangeB, andedRanges, doChecks)

    if (size(andedRanges) .ne. 2) then
      write(errstr, *) "Inconsistent andedRangesSize = ", size(andedRanges)
      call error("InconsistentAndedRangesVecSize1")
    else if (.not. areRangesEqual(andedRanges, rangeR)) then
      call dumpRanges(andedRanges)
      ! write(errstr, *) "Inconsistent andedRanges = " 
      call error("InconsistentAndedRangesVec1")
    end if

    deallocate(andedRanges)


    allocate(andedRanges(2))
    andedRanges = rangeB

    call andRanges(rangeA, andedRanges, doChecks)	! call andRangesVecVec(rangeA, andedRanges, doChecks)

    if (size(andedRanges) .ne. 2) then
      write(errstr, *) "Inconsistent andedRangesSize = ", size(andedRanges)
      call error("InconsistentAndedRangesVecSize2")
    else if (.not. areRangesEqual(andedRanges, rangeR)) then
      call dumpRanges(andedRanges)
      ! write(errstr, *) "Inconsistent andedRanges = " 
      call error("InconsistentAndedRangesVec2")
    end if

    deallocate(andedRanges)

  end subroutine test_andRangesVecVec


end program testdssrangesaux
