! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
program testtestutils
  use errorhandling
  use types
  use test_utils

  implicit none

  !! Testing all the interfaces of isNearlyEqual()

  ! Between Integers
  call isNEss	! Int8,   Int8
  call isNEsi	! Int8,   Int16
  call isNEii	! Int16,  Int16
  call isNEis	! Int16,  Int8
  call isNEsl	! Int8,   Int32
  call isNEil
  call isNEll
  call isNEli
  call isNEls

  ! Between Float/Integers
  call isNEsr	! Int8,   Single
  call isNEir	! Int16,  Single
  call isNElr	! Int32,  Single
  call isNErr	! Single, Single
  call isNErl	! Single, Int32
  call isNEri
  call isNErs

  call isNEsd	! Int8,   Double
  call isNEid	! Int16,  Double
  call isNEld
  call isNErd
  call isNEdd	! Double, Double
  call isNEdr
  call isNEdl
  call isNEdi
  call isNEds

  ! In document, \label{ssclib:description:test_utils:isNearlyEqual}
  if (.not. isNearlyEqual(1110,   1112,   3)) call error("anotherErr01")
  if (      isNearlyEqual(1.2,    1.0,    3)) call error("anotherErr02")
  if (.not. isNearlyEqual(1.234,  1.231,  3)) call error("anotherErr03")
 !if (      isNearlyEqual(1.2349, 1.2351, 3)) call error("anotherErr04")	! This used to be the case before Ver.4.7.
  if (.not. isNearlyEqual(1.2349, 1.2351, 3)) call error("anotherErr04")	! This is the case in Ver.4.7 and after.

contains

  subroutine isNEss
    integer(int8) :: a, b, c, d, z
    a=int(12.1d1, kind=int8)
    b=int(1.23d2, kind=int8)
    z=int(0,        kind=int8)
    c=int(1.23d1,   kind=int8)
    d=int(1.23d0,   kind=int8)

    if (.not. isNearlyEqual( a,  b, 2)) call error("isNEss01")
    if (.not. isNearlyEqual( b,  a, 2)) call error("isNEss03")	! Swap
    if (      isNearlyEqual( a,  b, 5)) call error("isNEss05")	! TooMuchPrec
    if (.not. isNearlyEqual(-a, -b, 2)) call error("isNEss07")	! Negative
    if (      isNearlyEqual(-a,  b, 2)) call error("isNEss09")	! Nega/Positive
    if (      isNearlyEqual( a, -b, 2)) call error("isNEss11")
    if (.not. isNearlyEqual( z,  z, 1)) call error("isNEss13")	! 0/0
    if (      isNearlyEqual( z,  c, 2)) call error("isNEss15")	! 0/non-0
    if (      isNearlyEqual( d,  z, 2)) call error("isNEss17")
    if (      isNearlyEqual( d,  b, 5)) call error("isNEss20")
  end subroutine isNEss

  subroutine isNEsi
    integer(int8)  :: a0, b0, c0, d0, z0
    integer(int16) :: a1, b1, c1, d1, z1
    a0=int(12.1d1, kind=int8)
    b0=int(1.23d2, kind=int8)
    z0=int(0,      kind=int8)
    c0=int(1.23d1, kind=int8)
    d0=int(1.23d0, kind=int8)
    a1=int(12.1d1, kind=int16)
    b1=int(1.23d2, kind=int16)
    z1=int(0,      kind=int16)
    c1=int(1.23d1, kind=int16)
    d1=int(1.23d0, kind=int16)

    if (.not. isNearlyEqual( a0,  b1, 2)) call error("isNEsi01")
    if (.not. isNearlyEqual( b0,  a1, 2)) call error("isNEsi03")
    if (      isNearlyEqual( a0,  b1, 5)) call error("isNEsi05")
    if (.not. isNearlyEqual(-a0, -b1, 2)) call error("isNEsi07")
    if (      isNearlyEqual(-a0,  b1, 2)) call error("isNEsi09")
    if (      isNearlyEqual( a0, -b1, 2)) call error("isNEsi11")
    if (.not. isNearlyEqual( z0,  z1, 1)) call error("isNEsi13")
    if (      isNearlyEqual( z0,  c1, 2)) call error("isNEsi15")
    if (      isNearlyEqual( d0,  z1, 2)) call error("isNEsi17")
  end subroutine isNEsi

  subroutine isNEii
    integer(int16) :: a, b, c, d, z
    a=int(12.346d3, kind=int16)
    b=int(1.2349d4, kind=int16)
    z=int(0,        kind=int16)
    c=int(1.23d1,   kind=int16)
    d=int(1.23d0,   kind=int16)

    if (.not. isNearlyEqual( a,  b, 4)) call error("isNEii01")
    if (.not. isNearlyEqual( b,  a, 4)) call error("isNEii03")
    if (      isNearlyEqual( a,  b, 5)) call error("isNEii05")
    if (.not. isNearlyEqual(-a, -b, 4)) call error("isNEii07")
    if (      isNearlyEqual(-a,  b, 4)) call error("isNEii09")
    if (      isNearlyEqual( a, -b, 4)) call error("isNEii11")
    if (.not. isNearlyEqual( z,  z, 1)) call error("isNEii13")
    if (      isNearlyEqual( z,  c, 2)) call error("isNEii15")
    if (      isNearlyEqual( d,  z, 2)) call error("isNEii17")
    if (      isNearlyEqual( d,  b, 5)) call error("isNEii20")
  end subroutine isNEii

  subroutine isNEis
    integer(int16) :: a0, b0, c0, d0, z0
    integer(int8)  :: a1, b1, c1, d1, z1
    a0=int(12.1d1, kind=int16)
    b0=int(1.23d2, kind=int16)
    z0=int(0,      kind=int16)
    c0=int(1.23d1, kind=int16)
    d0=int(1.23d0, kind=int16)
    a1=int(12.1d1, kind=int8)
    b1=int(1.23d2, kind=int8)
    z1=int(0,      kind=int8)
    c1=int(1.23d1, kind=int8)
    d1=int(1.23d0, kind=int8)

    if (.not. isNearlyEqual( a0,  b1, 2)) call error("isNEsi01")
    if (.not. isNearlyEqual( b0,  a1, 2)) call error("isNEsi03")
    if (      isNearlyEqual( a0,  b1, 5)) call error("isNEsi05")
    if (.not. isNearlyEqual(-a0, -b1, 2)) call error("isNEsi07")
    if (      isNearlyEqual(-a0,  b1, 2)) call error("isNEsi09")
    if (      isNearlyEqual( a0, -b1, 2)) call error("isNEsi11")
    if (.not. isNearlyEqual( z0,  z1, 1)) call error("isNEsi13")
    if (      isNearlyEqual( z0,  c1, 2)) call error("isNEsi15")
    if (      isNearlyEqual( d0,  z1, 2)) call error("isNEsi17")
  end subroutine isNEis

  subroutine isNEsl
    integer(int8)  :: a0, b0, c0, d0, z0
    integer(int32) :: a1, b1, c1, d1, z1
    a0=int(12.1d1, kind=int8)
    b0=int(1.23d2, kind=int8)
    z0=int(0,      kind=int8)
    c0=int(1.23d1, kind=int8)
    d0=int(1.23d0, kind=int8)
    a1=int(12.1d1, kind=int32)
    b1=int(1.23d2, kind=int32)
    z1=int(0,      kind=int32)
    c1=int(1.23d1, kind=int32)
    d1=int(1.23d0, kind=int32)

    if (.not. isNearlyEqual( a0,  b1, 2)) call error("isNEsl01")
    if (.not. isNearlyEqual( b0,  a1, 2)) call error("isNEsl03")
    if (      isNearlyEqual( a0,  b1, 5)) call error("isNEsl05")
    if (.not. isNearlyEqual(-a0, -b1, 2)) call error("isNEsl07")
    if (      isNearlyEqual(-a0,  b1, 2)) call error("isNEsl09")
    if (      isNearlyEqual( a0, -b1, 2)) call error("isNEsl11")
    if (.not. isNearlyEqual( z0,  z1, 1)) call error("isNEsl13")
    if (      isNearlyEqual( z0,  c1, 2)) call error("isNEsl15")
    if (      isNearlyEqual( d0,  z1, 2)) call error("isNEsl17")
  end subroutine isNEsl

  subroutine isNEil
    integer(int16) :: a0, b0, c0, d0, z0
    integer(int32) :: a1, b1, c1, d1, z1
    a0=int(12.346d3, kind=int16)
    b0=int(1.2349d4, kind=int16)
    z0=int(0,        kind=int16)
    c0=int(1.23d1,   kind=int16)
    d0=int(1.23d0,   kind=int16)
    a1=int(12.346d3, kind=int32)
    b1=int(1.2349d4, kind=int32)
    z1=int(0,        kind=int32)
    c1=int(1.23d1,   kind=int32)
    d1=int(1.23d0,   kind=int32)

    if (.not. isNearlyEqual( a0,  b1, 4)) call error("isNEil01")
    if (.not. isNearlyEqual( b0,  a1, 4)) call error("isNEil03")
    if (      isNearlyEqual( a0,  b1, 5)) call error("isNEil05")
    if (.not. isNearlyEqual(-a0, -b1, 4)) call error("isNEil07")
    if (      isNearlyEqual(-a0,  b1, 4)) call error("isNEil09")
    if (      isNearlyEqual( a0, -b1, 4)) call error("isNEil11")
    if (.not. isNearlyEqual( z0,  z1, 1)) call error("isNEil13")
    if (      isNearlyEqual( z0,  c1, 2)) call error("isNEil15")
    if (      isNearlyEqual( d0,  z1, 2)) call error("isNEil17")
  end subroutine isNEil

  subroutine isNEll
    integer(int32) :: a, b, c, d, z, f
    a=int(12.346d4, kind=int32)
    b=int(1.2349d5, kind=int32)
    z=int(0,        kind=int32)
    c=int(1.23d1,   kind=int32)
    d=int(1.23d0,   kind=int32)
    f=int(1.2349d8, kind=int32)

    if (.not. isNearlyEqual( a,  b, 4)) call error("isNEll01")
    if (.not. isNearlyEqual( b,  a, 4)) call error("isNEll03")
    if (      isNearlyEqual( a,  b, 5)) call error("isNEll05")
    if (.not. isNearlyEqual(-a, -b, 4)) call error("isNEll07")
    if (      isNearlyEqual(-a,  b, 4)) call error("isNEll09")
    if (      isNearlyEqual( a, -b, 4)) call error("isNEll11")
    if (.not. isNearlyEqual( z,  z, 1)) call error("isNEll13")
    if (      isNearlyEqual( z,  c, 2)) call error("isNEll15")
    if (      isNearlyEqual( d,  z, 2)) call error("isNEll17")
    if (      isNearlyEqual( d,  f, 5)) call error("isNEll20")
  end subroutine isNEll

  subroutine isNEli
    integer(int32) :: a0, b0, c0, d0, z0
    integer(int16) :: a1, b1, c1, d1, z1
    a0=int(12.346d3, kind=int32)
    b0=int(1.2349d4, kind=int32)
    z0=int(0,        kind=int32)
    c0=int(1.23d1,   kind=int32)
    d0=int(1.23d0,   kind=int32)
    a1=int(12.346d3, kind=int16)
    b1=int(1.2349d4, kind=int16)
    z1=int(0,        kind=int16)
    c1=int(1.23d1,   kind=int16)
    d1=int(1.23d0,   kind=int16)

    if (.not. isNearlyEqual( a0,  b1, 4)) call error("isNEli01")
    if (.not. isNearlyEqual( b0,  a1, 4)) call error("isNEli03")
    if (      isNearlyEqual( a0,  b1, 5)) call error("isNEli05")
    if (.not. isNearlyEqual(-a0, -b1, 4)) call error("isNEli07")
    if (      isNearlyEqual(-a0,  b1, 4)) call error("isNEli09")
    if (      isNearlyEqual( a0, -b1, 4)) call error("isNEli11")
    if (.not. isNearlyEqual( z0,  z1, 1)) call error("isNEli13")
    if (      isNearlyEqual( z0,  c1, 2)) call error("isNEli15")
    if (      isNearlyEqual( d0,  z1, 2)) call error("isNEli17")
  end subroutine isNEli

  subroutine isNEls
    integer(int32) :: a0, b0, c0, d0, z0
    integer(int8)  :: a1, b1, c1, d1, z1
    a0=int(12.1d1, kind=int32)
    b0=int(1.23d2, kind=int32)
    z0=int(0,      kind=int32)
    c0=int(1.23d1, kind=int32)
    d0=int(1.23d0, kind=int32)
    a1=int(12.1d1, kind=int8)
    b1=int(1.23d2, kind=int8)
    z1=int(0,      kind=int8)
    c1=int(1.23d1, kind=int8)
    d1=int(1.23d0, kind=int8)

    if (.not. isNearlyEqual( a0,  b1, 2)) call error("isNEls01")
    if (.not. isNearlyEqual( b0,  a1, 2)) call error("isNEls03")
    if (      isNearlyEqual( a0,  b1, 5)) call error("isNEls05")
    if (.not. isNearlyEqual(-a0, -b1, 2)) call error("isNEls07")
    if (      isNearlyEqual(-a0,  b1, 2)) call error("isNEls09")
    if (      isNearlyEqual( a0, -b1, 2)) call error("isNEls11")
    if (.not. isNearlyEqual( z0,  z1, 1)) call error("isNEls13")
    if (      isNearlyEqual( z0,  c1, 2)) call error("isNEls15")
    if (      isNearlyEqual( d0,  z1, 2)) call error("isNEls17")
  end subroutine isNEls

  subroutine isNEsr
    integer(int8)  :: a0, b0, c0, d0, z0
    real(single)   :: a1, b1, c1, d1, z1
    a0=int(12.1d1, kind=int8)
    b0=int(1.23d2, kind=int8)
    z0=int(0,      kind=int8)
    c0=int(1.23d1, kind=int8)
    d0=int(1.23d0, kind=int8)
    a1=12.1e1
    b1=1.23e2
    z1=0.0
    c1=1.23e1
    d1=1.23e0

    if (.not. isNearlyEqual( a0,  b1, 2)) call error("isNEsr01")
    if (.not. isNearlyEqual( b0,  a1, 2)) call error("isNEsr03")
    if (      isNearlyEqual( a0,  b1, 5)) call error("isNEsr05")
    if (.not. isNearlyEqual(-a0, -b1, 2)) call error("isNEsr07")
    if (      isNearlyEqual(-a0,  b1, 2)) call error("isNEsr09")
    if (      isNearlyEqual( a0, -b1, 2)) call error("isNEsr11")
    if (.not. isNearlyEqual( z0,  z1, 1)) call error("isNEsr13")
    if (      isNearlyEqual( z0,  c1, 2)) call error("isNEsr15")
    if (      isNearlyEqual( d0,  z1, 2)) call error("isNEsr17")
  end subroutine isNEsr

  subroutine isNEir
    integer(int16) :: a0, b0, c0, d0, z0
    real(single)   :: a1, b1, c1, d1, z1
    a0=int(12.346d3, kind=int16)
    b0=int(1.2349d4, kind=int16)
    z0=int(0,        kind=int16)
    c0=int(1.23d1,   kind=int16)
    d0=int(1.23d0,   kind=int16)
    a1=12.346e3
    b1=1.2349e4
    z1=0.0
    c1=1.23e1
    d1=1.23e0

    if (.not. isNearlyEqual( a0,  b1, 4)) call error("isNEir01")
    if (.not. isNearlyEqual( b0,  a1, 4)) call error("isNEir03")
    if (      isNearlyEqual( a0,  b1, 5)) call error("isNEir05")
    if (.not. isNearlyEqual(-a0, -b1, 4)) call error("isNEir07")
    if (      isNearlyEqual(-a0,  b1, 4)) call error("isNEir09")
    if (      isNearlyEqual( a0, -b1, 4)) call error("isNEir11")
    if (.not. isNearlyEqual( z0,  z1, 1)) call error("isNEir13")
    if (      isNearlyEqual( z0,  c1, 2)) call error("isNEir15")
    if (      isNearlyEqual( d0,  z1, 2)) call error("isNEir17")
  end subroutine isNEir

  subroutine isNElr
    integer(int32) :: a0, b0, c0, d0, z0
    real(single)   :: a1, b1, c1, d1, z1
    a0=int(12.346d3, kind=int32)
    b0=int(1.2349d4, kind=int32)
    z0=int(0,        kind=int32)
    c0=int(1.23d1,   kind=int32)
    d0=int(1.23d0,   kind=int32)
    a1=12.346e3
    b1=1.2349e4
    z1=0.0
    c1=1.23e1
    d1=1.23e0

    if (.not. isNearlyEqual( a0,  b1, 4)) call error("isNElr01")
    if (.not. isNearlyEqual( b0,  a1, 4)) call error("isNElr03")
    if (      isNearlyEqual( a0,  b1, 5)) call error("isNElr05")
    if (.not. isNearlyEqual(-a0, -b1, 4)) call error("isNElr07")
    if (      isNearlyEqual(-a0,  b1, 4)) call error("isNElr09")
    if (      isNearlyEqual( a0, -b1, 4)) call error("isNElr11")
    if (.not. isNearlyEqual( z0,  z1, 1)) call error("isNElr13")
    if (      isNearlyEqual( z0,  c1, 2)) call error("isNElr15")
    if (      isNearlyEqual( d0,  z1, 2)) call error("isNElr17")
  end subroutine isNElr

  subroutine isNErr
    if (.not. isNearlyEqual( 12.346e0,  1.2349e1, 4)) call error("isNErr01")
    if (.not. isNearlyEqual( 1.2349e1,  12.346e0, 4)) call error("isNErr03")
    if (      isNearlyEqual( 12.346e0,  1.2349e1, 5)) call error("isNErr05")
    if (.not. isNearlyEqual(-12.346e0, -1.2349e1, 4)) call error("isNErr07")
    if (      isNearlyEqual(-12.346e0,  1.2349e1, 4)) call error("isNErr09")
    if (      isNearlyEqual( 12.346e0, -1.2349e1, 4)) call error("isNErr11")
    if (.not. isNearlyEqual(    0.0e0,     0.0e1, 2)) call error("isNErr13")
    if (      isNearlyEqual(    0.0e0,    1.23e1, 2)) call error("isNErr15")
    if (      isNearlyEqual(   1.23e0,     0.0e1, 2)) call error("isNErr17")
    if (      isNearlyEqual(   1.23e0,    3.4e10, 2)) call error("isNErr20")	! used to cause "Arithmetic exception" before Ver.4.7
  end subroutine isNErr

  subroutine isNErl
    real(single)   :: a0, b0, c0, d0, z0
    integer(int32) :: a1, b1, c1, d1, z1
    a0=12.346e3
    b0=1.2349e4
    z0=0.0e0
    c0=1.23e1
    d0=1.23e0
    a1=int(12.346d3, kind=int32)
    b1=int(1.2349d4, kind=int32)
    z1=int(0,        kind=int32)
    c1=int(1.23d1,   kind=int32)
    d1=int(1.23d0,   kind=int32)

    if (.not. isNearlyEqual( a0,  b1, 4)) call error("isNErl01")
    if (.not. isNearlyEqual( b0,  a1, 4)) call error("isNErl03")
    if (      isNearlyEqual( a0,  b1, 5)) call error("isNErl05")
    if (.not. isNearlyEqual(-a0, -b1, 4)) call error("isNErl07")
    if (      isNearlyEqual(-a0,  b1, 4)) call error("isNErl09")
    if (      isNearlyEqual( a0, -b1, 4)) call error("isNErl11")
    if (.not. isNearlyEqual( z0,  z1, 1)) call error("isNErl13")
    if (      isNearlyEqual( z0,  c1, 2)) call error("isNErl15")
    if (      isNearlyEqual( d0,  z1, 2)) call error("isNErl17")
  end subroutine isNErl

  subroutine isNEri
    real(single)   :: a0, b0, c0, d0, z0
    integer(int16) :: a1, b1, c1, d1, z1
    a0=12.346e3
    b0=1.2349e4
    z0=0.0e0
    c0=1.23e1
    d0=1.23e0
    a1=int(12.346d3, kind=int16)
    b1=int(1.2349d4, kind=int16)
    z1=int(0,        kind=int16)
    c1=int(1.23d1,   kind=int16)
    d1=int(1.23d0,   kind=int16)

    if (.not. isNearlyEqual( a0,  b1, 4)) call error("isNEri01")
    if (.not. isNearlyEqual( b0,  a1, 4)) call error("isNEri03")
    if (      isNearlyEqual( a0,  b1, 5)) call error("isNEri05")
    if (.not. isNearlyEqual(-a0, -b1, 4)) call error("isNEri07")
    if (      isNearlyEqual(-a0,  b1, 4)) call error("isNEri09")
    if (      isNearlyEqual( a0, -b1, 4)) call error("isNEri11")
    if (.not. isNearlyEqual( z0,  z1, 1)) call error("isNEri13")
    if (      isNearlyEqual( z0,  c1, 2)) call error("isNEri15")
    if (      isNearlyEqual( d0,  z1, 2)) call error("isNEri17")
  end subroutine isNEri

  subroutine isNErs
    real(single)  :: a0, b0, c0, d0, z0
    integer(int8) :: a1, b1, c1, d1, z1
    a0=12.1e1
    b0=1.23e2
    z0=0.0
    c0=1.23e1
    d0=1.23e0
    a1=int(12.1d1, kind=int32)
    b1=int(1.23d2, kind=int32)
    z1=int(0,      kind=int32)
    c1=int(1.23d1, kind=int32)
    d1=int(1.23d0, kind=int32)

    if (.not. isNearlyEqual( a0,  b1, 2)) call error("isNErs01")
    if (.not. isNearlyEqual( b0,  a1, 2)) call error("isNErs03")
    if (      isNearlyEqual( a0,  b1, 5)) call error("isNErs05")
    if (.not. isNearlyEqual(-a0, -b1, 2)) call error("isNErs07")
    if (      isNearlyEqual(-a0,  b1, 2)) call error("isNErs09")
    if (      isNearlyEqual( a0, -b1, 2)) call error("isNErs11")
    if (.not. isNearlyEqual( z0,  z1, 1)) call error("isNErs13")
    if (      isNearlyEqual( z0,  c1, 2)) call error("isNErs15")
    if (      isNearlyEqual( d0,  z1, 2)) call error("isNErs17")
  end subroutine isNErs

  subroutine isNEsd
    integer(int8)  :: a0, b0, c0, d0, z0
    real(double)   :: a1, b1, c1, d1, z1
    a0=int(12.1d1, kind=int8)
    b0=int(1.23d2, kind=int8)
    z0=int(0,      kind=int8)
    c0=int(1.23d1, kind=int8)
    d0=int(1.23d0, kind=int8)
    a1=12.1d1
    b1=1.23d2
    z1=0.0d0
    c1=1.23d1
    d1=1.23d0

    if (.not. isNearlyEqual( a0,  b1, 2)) call error("isNEsd01")
    if (.not. isNearlyEqual( b0,  a1, 2)) call error("isNEsd03")
    if (      isNearlyEqual( a0,  b1, 5)) call error("isNEsd05")
    if (.not. isNearlyEqual(-a0, -b1, 2)) call error("isNEsd07")
    if (      isNearlyEqual(-a0,  b1, 2)) call error("isNEsd09")
    if (      isNearlyEqual( a0, -b1, 2)) call error("isNEsd11")
    if (.not. isNearlyEqual( z0,  z1, 1)) call error("isNEsd13")
    if (      isNearlyEqual( z0,  c1, 2)) call error("isNEsd15")
    if (      isNearlyEqual( d0,  z1, 2)) call error("isNEsd17")
  end subroutine isNEsd

  subroutine isNEid
    integer(int16) :: a0, b0, c0, d0, z0
    real(double)   :: a1, b1, c1, d1, z1
    a0=int(12.346d3, kind=int16)
    b0=int(1.2349d4, kind=int16)
    z0=int(0,        kind=int16)
    c0=int(1.23d1,   kind=int16)
    d0=int(1.23d0,   kind=int16)
    a1=12.346d3
    b1=1.2349d4
    z1=0.0d0
    c1=1.23d1
    d1=1.23d0

    if (.not. isNearlyEqual( a0,  b1, 4)) call error("isNEid01")
    if (.not. isNearlyEqual( b0,  a1, 4)) call error("isNEid03")
    if (      isNearlyEqual( a0,  b1, 5)) call error("isNEid05")
    if (.not. isNearlyEqual(-a0, -b1, 4)) call error("isNEid07")
    if (      isNearlyEqual(-a0,  b1, 4)) call error("isNEid09")
    if (      isNearlyEqual( a0, -b1, 4)) call error("isNEid11")
    if (.not. isNearlyEqual( z0,  z1, 1)) call error("isNEid13")
    if (      isNearlyEqual( z0,  c1, 2)) call error("isNEid15")
    if (      isNearlyEqual( d0,  z1, 2)) call error("isNEid17")
  end subroutine isNEid

  subroutine isNEld
    integer(int32) :: a0, b0, c0, d0, z0
    real(double)   :: a1, b1, c1, d1, z1
    a0=int(12.346d3, kind=int32)
    b0=int(1.2349d4, kind=int32)
    z0=int(0,        kind=int32)
    c0=int(1.23d1,   kind=int32)
    d0=int(1.23d0,   kind=int32)
    a1=12.346d3
    b1=1.2349d4
    z1=0.0d0
    c1=1.23d1
    d1=1.23d0

    if (.not. isNearlyEqual( a0,  b1, 4)) call error("isNEld01")
    if (.not. isNearlyEqual( b0,  a1, 4)) call error("isNEld03")
    if (      isNearlyEqual( a0,  b1, 5)) call error("isNEld05")
    if (.not. isNearlyEqual(-a0, -b1, 4)) call error("isNEld07")
    if (      isNearlyEqual(-a0,  b1, 4)) call error("isNEld09")
    if (      isNearlyEqual( a0, -b1, 4)) call error("isNEld11")
    if (.not. isNearlyEqual( z0,  z1, 1)) call error("isNEld13")
    if (      isNearlyEqual( z0,  c1, 2)) call error("isNEld15")
    if (      isNearlyEqual( d0,  z1, 2)) call error("isNEld17")
  end subroutine isNEld

  subroutine isNErd
    if (.not. isNearlyEqual( 12.346e0,  1.2349d1, 4)) call error("isNErd01")
    if (.not. isNearlyEqual( 1.2349e1,  12.346d0, 4)) call error("isNErd03")
    if (      isNearlyEqual( 12.346e0,  1.2349d1, 5)) call error("isNErd05")
    if (.not. isNearlyEqual(-12.346e0, -1.2349d1, 4)) call error("isNErd07")
    if (      isNearlyEqual(-12.346e0,  1.2349d1, 4)) call error("isNErd09")
    if (      isNearlyEqual( 12.346e0, -1.2349d1, 4)) call error("isNErd11")
    if (.not. isNearlyEqual(    0.0e0,     0.0d1, 2)) call error("isNErd13")
    if (      isNearlyEqual(    0.0e0,    1.23d1, 2)) call error("isNErd15")
    if (      isNearlyEqual(   1.23e0,     0.0d1, 2)) call error("isNErd17")
  end subroutine isNErd

  subroutine isNEdd
    if (.not. isNearlyEqual( 12.346d0,  1.2349d1, 4)) call error("isNEdd01")
    if (.not. isNearlyEqual( 1.2349d1,  12.346d0, 4)) call error("isNEdd03")
    if (      isNearlyEqual( 12.346d0,  1.2349d1, 5)) call error("isNEdd05")
    if (.not. isNearlyEqual(-12.346d0, -1.2349d1, 4)) call error("isNEdd07")
    if (      isNearlyEqual(-12.346d0,  1.2349d1, 4)) call error("isNEdd09")
    if (      isNearlyEqual( 12.346d0, -1.2349d1, 4)) call error("isNEdd11")
    if (.not. isNearlyEqual(    0.0d0,     0.0d1, 2)) call error("isNEdd13")
    if (      isNearlyEqual(    0.0d0,    1.23d1, 2)) call error("isNEdd15")
    if (      isNearlyEqual(   1.23d0,     0.0d1, 2)) call error("isNEdd17")
    if (      isNearlyEqual(   1.23d0,    3.4d10, 2)) call error("isNEdd20")	! used to cause "Arithmetic exception" before Ver.4.7
  end subroutine isNEdd

  subroutine isNEdr
    if (.not. isNearlyEqual( 12.346d0,  1.2349e1, 4)) call error("isNEdr01")
    if (.not. isNearlyEqual( 1.2349d1,  12.346e0, 4)) call error("isNEdr03")
    if (      isNearlyEqual( 12.346d0,  1.2349e1, 5)) call error("isNEdr05")
    if (.not. isNearlyEqual(-12.346d0, -1.2349e1, 4)) call error("isNEdr07")
    if (      isNearlyEqual(-12.346d0,  1.2349e1, 4)) call error("isNEdr09")
    if (      isNearlyEqual( 12.346d0, -1.2349e1, 4)) call error("isNEdr11")
    if (.not. isNearlyEqual(    0.0d0,     0.0e1, 2)) call error("isNEdr13")
    if (      isNearlyEqual(    0.0d0,    1.23e1, 2)) call error("isNEdr15")
    if (      isNearlyEqual(   1.23d0,     0.0e1, 2)) call error("isNEdr17")
  end subroutine isNEdr

  subroutine isNEdl
    real(double)   :: a0, b0, c0, d0, z0
    integer(int32) :: a1, b1, c1, d1, z1
    a0=12.346d3
    b0=1.2349d4
    z0=0.0d0
    c0=1.23d1
    d0=1.23d0
    a1=int(12.346d3, kind=int32)
    b1=int(1.2349d4, kind=int32)
    z1=int(0,        kind=int32)
    c1=int(1.23d1,   kind=int32)
    d1=int(1.23d0,   kind=int32)

    if (.not. isNearlyEqual( a0,  b1, 4)) call error("isNEdl01")
    if (.not. isNearlyEqual( b0,  a1, 4)) call error("isNEdl03")
    if (      isNearlyEqual( a0,  b1, 5)) call error("isNEdl05")
    if (.not. isNearlyEqual(-a0, -b1, 4)) call error("isNEdl07")
    if (      isNearlyEqual(-a0,  b1, 4)) call error("isNEdl09")
    if (      isNearlyEqual( a0, -b1, 4)) call error("isNEdl11")
    if (.not. isNearlyEqual( z0,  z1, 1)) call error("isNEdl13")
    if (      isNearlyEqual( z0,  c1, 2)) call error("isNEdl15")
    if (      isNearlyEqual( d0,  z1, 2)) call error("isNEdl17")
  end subroutine isNEdl

  subroutine isNEdi
    real(double)   :: a0, b0, c0, d0, z0
    integer(int16) :: a1, b1, c1, d1, z1
    a0=12.346d3
    b0=1.2349d4
    z0=0.0d0
    c0=1.23d1
    d0=1.23d0
    a1=int(12.346d3, kind=int16)
    b1=int(1.2349d4, kind=int16)
    z1=int(0,        kind=int16)
    c1=int(1.23d1,   kind=int16)
    d1=int(1.23d0,   kind=int16)

    if (.not. isNearlyEqual( a0,  b1, 4)) call error("isNEdi01")
    if (.not. isNearlyEqual( b0,  a1, 4)) call error("isNEdi03")
    if (      isNearlyEqual( a0,  b1, 5)) call error("isNEdi05")
    if (.not. isNearlyEqual(-a0, -b1, 4)) call error("isNEdi07")
    if (      isNearlyEqual(-a0,  b1, 4)) call error("isNEdi09")
    if (      isNearlyEqual( a0, -b1, 4)) call error("isNEdi11")
    if (.not. isNearlyEqual( z0,  z1, 1)) call error("isNEdi13")
    if (      isNearlyEqual( z0,  c1, 2)) call error("isNEdi15")
    if (      isNearlyEqual( d0,  z1, 2)) call error("isNEdi17")
  end subroutine isNEdi

  subroutine isNEds
    real(double)  :: a0, b0, c0, d0, z0
    integer(int8) :: a1, b1, c1, d1, z1
    a0=12.1d1
    b0=1.23d2
    z0=0.0d0
    c0=1.23d1
    d0=1.23d0
    a1=int(12.1d1, kind=int32)
    b1=int(1.23d2, kind=int32)
    z1=int(0,      kind=int32)
    c1=int(1.23d1, kind=int32)
    d1=int(1.23d0, kind=int32)

    if (.not. isNearlyEqual( a0,  b1, 2)) call error("isNEds01")
    if (.not. isNearlyEqual( b0,  a1, 2)) call error("isNEds03")
    if (      isNearlyEqual( a0,  b1, 5)) call error("isNEds05")
    if (.not. isNearlyEqual(-a0, -b1, 2)) call error("isNEds07")
    if (      isNearlyEqual(-a0,  b1, 2)) call error("isNEds09")
    if (      isNearlyEqual( a0, -b1, 2)) call error("isNEds11")
    if (.not. isNearlyEqual( z0,  z1, 1)) call error("isNEds13")
    if (      isNearlyEqual( z0,  c1, 2)) call error("isNEds15")
    if (      isNearlyEqual( d0,  z1, 2)) call error("isNEds17")
  end subroutine isNEds

end program testtestutils
