! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
program testwcsaux
  use errorhandling
  use types
  use dal		! For INTEGER_NULL
  ! use cal

  !! Followings are in ssclib
  use array_utils	! for aryStatInfoSingleT and getAryStatInfo()
  !use cal_aux		! for getCalInfo()
  use dal_aux		! for readArrayData()
  use constants		! for Pi
  !use coordinate
  use test_utils	! for isNearlyEqual()
  use wcs_aux

  implicit none

  call test_writeWcsExtendedArray

contains

  !---------------------------------------------------------------------
  subroutine test_writeWcsExtendedArray
  !---------------------------------------------------------------------
  ! Tests of writeWcsExtendedArray() (interfaced from writeWcs) in wcs_aux
  !

    implicit none

    character(30)  :: errHead
    character(*), parameter :: DefOutFilename = 'outfiles/wcsOut001a.ds'
    character(*), parameter :: DefTblName = 'some_table'

!     real(single), allocatable :: arySgl(:,:)
!     real(double), allocatable :: aryDbl(:,:)
!     integer(int32), allocatable :: aryInt32(:,:)
!     integer(int16), allocatable :: aryInt16(:,:)
!     integer(int8),  allocatable :: aryInt8(:,:)
!     type(AryEdgesInfoT) :: aryEdgesInfo	! defined in array_utils
!     ! real(single) :: offset
!     character(256) :: outFilename, refFilename
!     character(64)  :: outObs_id, refObs_id, outInstrument, refInstrument, outTelescop
!     logical :: flagExist
!     type(DataSetT) :: detSet, refSet
!     type(ArrayT) :: detArray
!     type(WcsT) :: refWcs			! defined in wcs_aux
!     type(WcsAxesExtendedT) :: wcsExtended	! defined in wcs_aux
!     real(single), pointer :: arySglRead(:,:)
!     real(double), pointer :: aryDblRead(:,:)
!     integer(int32), pointer :: aryInt32Read(:,:)
!     integer(int16), pointer :: aryInt16Read(:,:)
!     integer(int8),  pointer :: aryInt8Read(:,:)
!     real(double), dimension(2) :: crvall, cdeltl, crval
!     integer,      dimension(2) :: crpixl

    type(DataSetT) :: setIn, setOut
    type(ArrayT)   :: arrIn, arrOut
    type(WcsT) :: refWcs			! defined in wcs_aux
    type(WcsAxesExtendedT) :: wcsE, wcsE2
    real(double), parameter   :: DefDbl = 2.0d0
    integer(int32), parameter :: DefInt = 3

    !-------------------------------------------------------------------
    ! Start:

    !----------------------------------------------------------
    errHead = "WriteWcsExtendedArray001a"
    print *,"=== "//trim(errHead)//": Simple."

    ! Taken from /dal/tests/f90test/createarray.f90
    setOut = dataSet(DefOutFilename, CREATE)
    arrOut = addArray(setOut, DefTblName, INTEGER32, (/ 3,4,2 /), units="mm" )
    call release(setOut)

    ! refFilename = 'outfiles/pn_2000.ds'	! Should have been prepared by another executable in this test harness!
    call readWcs('outfiles/pn_2000.ds', refWcs)	! Parameterise in the header!!
    call convertWcsToExtended(refWcs, wcsE)	! defined in wcs_aux
    wcsE%CRPIXL = (/ 77, 77 /)
    wcsE%CRVALL(2) = -777.77d0
    wcsE%withPhysical = .true.

    setOut = dataSet(DefOutFilename, MODIFY)
    arrOut = array(setOut, DefTblName, MODIFY)
    wcsE%CRPIX(1)  = DefDbl
    wcsE%CRPIXL(2) = DefInt
    !wcsE%LTV1 = 0.0/(wcsE%LTV1-wcsE%LTV1)
    !wcsE%CRVAL(2)  = real64Attribute(setOut, 'Naiyo')

call dumpWcsExtended(wcsE, verbosityArg=0_int32, varName='WriteWcsExtendedArray001a')	! => (LTV2 == NaN)
    !call dumpWcsExtended(wcsE)
    call writeWcs(arrOut, wcsE)
    call release(setOut)
print *,'DEBUG:end...'

    call readWcs(DefOutFilename, refWcs)	! Parameterise in the header!!
    call convertWcsToExtended(refWcs, wcsE2)	! defined in wcs_aux

    setIn = dataSet(DefOutFilename, READ)
    arrIn = array(setIn, 0, READ)

    !! Check header attirbutes that should exist.
    if (wcsE%CRPIX(1) .ne. wcsE2%CRPIX(1)) then
    ! if (.not. isNearlyEqual(wcsE%CRPIX(1), wcsE2%CRPIX(1), 6)	! defined in test_utils
      print *,'wcsE2%CRPIX(1)=', wcsE2%CRPIX(1)
      call error("err"//trim(errHead), 'Inconsistent CRPIX')
    ! else if (wcsE%CRVALL(2) .ne. wcsE2%CRVALL(2)) then
    else if (wcsE%CRVALL(2) .ne. real64Attribute(arrIn, 'CRVAL2L')) then
      print *,'wcsE2%CRVALL(2)=', wcsE2%CRVALL(2)
      call error("err"//trim(errHead), 'Inconsistent CRVALL')
    end if

    !! Check header attirbutes that should or should not exist.
    if( .not. hasAttribute( arrIn, "CDELT2" )) then
      call error("err"//trim(errHead), 'No CDELT2 in ('//trim(DefOutFilename)//').')
    else if(hasAttribute(arrIn, "CRVAL1L" )) then
      call error("err"//trim(errHead), 'CRVAL1L should not exist in ('//trim(DefOutFilename)//').')
    end if

    call release(setIn)


!     allocate(arySgl(512,512))
! 
!     arySgl(:,:)   = 0.0
!     arySgl(256,:) = 1.0
!     outFilename = 'outfiles/detSgl_im001a.ds'
! 
!     call writeWcsExtendedArray(arySgl, trim(outFilename))
! 
!     ! Test whether the output file exists or not.
!     inquire(file=trim(outFilename), exist=flagExist)
!     if ( .not. flagExist) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') not exists.')
!     end if
! 
!     detSet = dataSet(outFilename, READ)
!     detArray = array(detSet, 0, READ)
! 
!     ! Tests for the hearder attributes.
!     !if (.not. hasAttribute(detSet, 'CRPIX1L') .or. &	! => Error
!     if (.not. hasAttribute(detArray, 'CRPIX1L') .or. &
!         .not. hasAttribute(detArray, 'CRVAL1L') .or. &
!         .not. hasAttribute(detArray, 'CDELT1L') .or. &
!         .not. hasAttribute(detArray, 'CTYPE1L') .or. &
!         .not. hasAttribute(detArray, 'CRPIX2L') .or. &
!         .not. hasAttribute(detArray, 'CRVAL2L') .or. &
!         .not. hasAttribute(detArray, 'CDELT2L') .or. &
!         .not. hasAttribute(detArray, 'CTYPE2L')) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') do not have required header attributes.')
!     end if
! 
!     crpixl(1) = int16Attribute( detArray, 'CRPIX1L')
!     crpixl(2) = int16Attribute( detArray, 'CRPIX2L')
!     crvall(1) = real64Attribute(detArray, 'CRVAL1L')
!     crvall(2) = real64Attribute(detArray, 'CRVAL2L')
!     cdeltl(1) = real64Attribute(detArray, 'CDELT1L')
!     cdeltl(2) = real64Attribute(detArray, 'CDELT2L')
!     if (crpixl(1) * crpixl(2) .ne. 1) then
!       print *,'CRPIXL=', crpixl
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong CRPIXL.')
!     else if ((nint(crvall(1)) .ne. -20440) .or. &	! calc '-512*80/2+80/2' - the latter is for half a pixel for the corresponding physical coordinates at CRPIXL.
!              (nint(crvall(2)) .ne. -20440)) then
!       print *,'CRVALL=', crvall
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong CRVALL.')
!     else if ((nint(cdeltl(1)) .ne. 80) .or. &
!              (nint(cdeltl(2)) .ne. 80)) then
!       print *,'CDELTL=', cdeltl
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong CDELTL.')
!     end if
! 
!     ! Tests for the image array itself.
!     call readArrayData(detSet, arySglRead)
!     ! print *,'lbound=',lbound(arySglRead)	! => (1,1)
!     if ( any(arySglRead(256,:) .ne. 1.0) .or. &
!          any(arySglRead(1:255,:) .ne. 0.0) ) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has wrong values.')
!     end if
!     deallocate(arySglRead)
!     call release(detSet)
!     deallocate(arySgl)
! 
! 
!     !----------------------------------------------------------
!     errHead = "SaveDetImage002a"
!     print *,"=== "//trim(errHead)//": With aryEdgesInfo (single)."
! 
!     allocate(arySgl(512,512))
! 
!     arySgl(:,:)   = 0.0
!     arySgl(256,:) = 1.0
!     outFilename = 'outfiles/detSgl_im002a.ds'
! 
!     aryEdgesInfo = getDetImageEdgesInfo(arySgl, outBinSizeXY=(/ 40.0, 40.0 /), outOffsetXY=(/10.0, 10.0/))
! 
!     call writeWcsExtendedArray(arySgl, trim(outFilename), detImageEdgesInfo=aryEdgesInfo )
! 
!     ! Test whether the output file exists or not.
!     inquire(file=trim(outFilename), exist=flagExist)
!     if ( .not. flagExist) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') not exists.')
!     end if
! 
!     detSet = dataSet(outFilename, READ)
!     detArray = array(detSet, 0, READ)
! 
!     ! Tests for the hearder attributes.
!     if (.not. hasAttribute(detArray, 'CRPIX1L') .or. &
!         .not. hasAttribute(detArray, 'CRVAL1L') .or. &
!         .not. hasAttribute(detArray, 'CDELT1L') .or. &
!         .not. hasAttribute(detArray, 'CTYPE1L') .or. &
!         .not. hasAttribute(detArray, 'CRPIX2L') .or. &
!         .not. hasAttribute(detArray, 'CRVAL2L') .or. &
!         .not. hasAttribute(detArray, 'CDELT2L') .or. &
!         .not. hasAttribute(detArray, 'CTYPE2L')) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') do not have required header attributes.')
!     end if
! 
!     crpixl(1) = int16Attribute( detArray, 'CRPIX1L')
!     crpixl(2) = int16Attribute( detArray, 'CRPIX2L')
!     crvall(1) = real64Attribute(detArray, 'CRVAL1L')
!     crvall(2) = real64Attribute(detArray, 'CRVAL2L')
!     cdeltl(1) = real64Attribute(detArray, 'CDELT1L')
!     cdeltl(2) = real64Attribute(detArray, 'CDELT2L')
!     if (crpixl(1) * crpixl(2) .ne. 1) then
!       print *,'CRPIXL=', crpixl
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong CRPIXL.')
!     else if ((nint(crvall(1)) .ne. -10210) .or. &	! calc '-512*40/2+40/2+10'
!              (nint(crvall(2)) .ne. -10210)) then
!       print *,'CRVALL=', crvall
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong CRVALL.')
!     else if ((nint(cdeltl(1)) .ne. 40) .or. &
!              (nint(cdeltl(2)) .ne. 40)) then
!       print *,'CDELTL=', cdeltl
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong CDELTL.')
!     end if
! 
!     ! Tests for the image array itself.
!     call readArrayData(detSet, arySglRead)
!     if ( any(arySglRead(256,:) .ne. 1.0) .or. &
!          any(arySglRead(1:255,:) .ne. 0.0) ) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has wrong values.')
!     end if
!     deallocate(arySglRead)
!     call release(detSet)
!     deallocate(arySgl)
! 
! 
!     !----------------------------------------------------------
!     errHead = "SaveDetImage003a"
!     print *,"=== "//trim(errHead)//": With templateFitsSetName for an event file (single)."
! 
!     allocate(arySgl(512,512))
! 
!     arySgl(:,:)   = 0.0
!     arySgl(256,:) = 1.0
!     outFilename = 'outfiles/detSgl_im003a.ds'
!     refFilename = 'outfiles/P0403760401M1S021MIEVLI0000.FIT'	! Should have been prepared by another executable in this test harness!
! 
!     refSet = dataSet(refFilename, READ)
!     refObs_id = ''
!     refInstrument = ''
!     refObs_id     = stringAttribute( refSet, 'OBS_ID')
!     refInstrument = stringAttribute( refSet, 'INSTRUME')
!     call release(refSet)
! 
!     call writeWcsExtendedArray(arySgl, trim(outFilename), templateFitsSetName=trim(refFilename))
! 
!     ! Test whether the output file exists or not.
!     inquire(file=trim(outFilename), exist=flagExist)
!     if ( .not. flagExist) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') not exists.')
!     end if
! 
!     detSet = dataSet(outFilename, READ)
!     detArray = array(detSet, 0, READ)
! 
!     ! Tests for the hearder attributes.
!     if (.not. hasAttribute(detArray, 'CRPIX1L') .or. &
!         .not. hasAttribute(detArray, 'CRVAL1L') .or. &
!         .not. hasAttribute(detArray, 'CDELT1L') .or. &
!         .not. hasAttribute(detArray, 'CTYPE1L') .or. &
!         .not. hasAttribute(detArray, 'CRPIX2L') .or. &
!         .not. hasAttribute(detArray, 'CRVAL2L') .or. &
!         .not. hasAttribute(detArray, 'CDELT2L') .or. &
!         .not. hasAttribute(detArray, 'CTYPE2L')) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') do not have required header attributes.')
!     end if
! 
!     crpixl(1) = int16Attribute( detArray, 'CRPIX1L')
!     crpixl(2) = int16Attribute( detArray, 'CRPIX2L')
!     crvall(1) = real64Attribute(detArray, 'CRVAL1L')
!     crvall(2) = real64Attribute(detArray, 'CRVAL2L')
!     cdeltl(1) = real64Attribute(detArray, 'CDELT1L')
!     cdeltl(2) = real64Attribute(detArray, 'CDELT2L')
!     if (crpixl(1) * crpixl(2) .ne. 1) then
!       print *,'CRPIXL=', crpixl
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong CRPIXL.')
!     else if ((nint(crvall(1)) .ne. -20440) .or. &	! calc '-512*80/2+80/2' - the latter is for half a pixel for the corresponding physical coordinates at CRPIXL.
!              (nint(crvall(2)) .ne. -20440)) then
!       print *,'CRVALL=', crvall
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong CRVALL.')
!     else if ((nint(cdeltl(1)) .ne. 80) .or. &
!              (nint(cdeltl(2)) .ne. 80)) then
!       print *,'CDELTL=', cdeltl
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong CDELTL.')
!     end if
! 
!     ! Tests for the hearder attributes inherited from the reference FITS file.
!     if (.not. hasAttribute(detSet, 'OBS_ID') .or. &
!         .not. hasAttribute(detSet, 'INSTRUME')) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has no OBS_ID or INSTRUME attribute.')
!     end if
! 
!     outObs_id = ''
!     outInstrument = ''
!     outObs_id     = stringAttribute( detSet, 'OBS_ID')
!     outInstrument = stringAttribute( detSet, 'INSTRUME')
! 
!     if (trim(outObs_id) .ne. trim(refObs_id)) then
!       print *,'outObs_id=', trim(outObs_id)
!       print *,'refObs_id=', trim(refObs_id)
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong OBS_ID.')
!     else if (trim(outInstrument) .ne. trim(refInstrument)) then
!       print *,'outInstrument=', trim(outInstrument)
!       print *,'refInstrument=', trim(refInstrument)
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong INSTRUME.')
!     end if
! 
!     ! Tests for the image array itself.
!     call readArrayData(detSet, arySglRead)
!     if ( any(arySglRead(256,:) .ne. 1.0) .or. &
!          any(arySglRead(1:255,:) .ne. 0.0) ) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has wrong values.')
!     end if
!     deallocate(arySglRead)
!     call release(detSet)
!     deallocate(arySgl)
! 
! 
!     !----------------------------------------------------------
!     errHead = "SaveDetImage004a"
!     print *,"=== "//trim(errHead)//": With templateFitsSetName for an image file and with detWcs (single)."
! 
!     allocate(arySgl(512,512))
! 
!     arySgl(:,:)   = 0.0
!     arySgl(256,:) = 1.0
!     outFilename = 'outfiles/detSgl_im004a.ds'
!     refFilename = 'outfiles/pn_2000.ds'	! Should have been prepared by another executable in this test harness!
! 
!     refSet = dataSet(refFilename, READ)
!     refObs_id = ''
!     refInstrument = ''
!     refObs_id     = stringAttribute( refSet, 'OBS_ID')
!     refInstrument = stringAttribute( refSet, 'INSTRUME')
! ! print *,'before readW'
! !     call readWcs(refSet, refWcs)	! There is a bug in ssclib/wcs_aux.f90/readWcs (for the interface to read DataSet) !!!!!
! ! print *,'after readW'
!     call release(refSet)
!     call readWcs(refFilename, refWcs)	! Move above, after readWcs is fixed.
! 
!     call writeWcsExtendedArray(arySgl, trim(outFilename), templateFitsSetName=trim(refFilename) &
!        , detWcs=refWcs)
! 
!     ! Test whether the output file exists or not.
!     inquire(file=trim(outFilename), exist=flagExist)
!     if ( .not. flagExist) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') not exists.')
!     end if
! 
!     detSet = dataSet(outFilename, READ)
!     detArray = array(detSet, 0, READ)
! 
!     ! Tests for the hearder attributes.
!     if (.not. hasAttribute(detArray, 'CRPIX1L') .or. &
!         .not. hasAttribute(detArray, 'CRVAL1L') .or. &
!         .not. hasAttribute(detArray, 'CDELT1L') .or. &
!         .not. hasAttribute(detArray, 'CTYPE1L') .or. &
!         .not. hasAttribute(detArray, 'CRPIX2L') .or. &
!         .not. hasAttribute(detArray, 'CRVAL2L') .or. &
!         .not. hasAttribute(detArray, 'CDELT2L') .or. &
!         .not. hasAttribute(detArray, 'CTYPE2L')) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') do not have required header attributes.')
!     end if
! 
!     if (hasAttribute(detArray, 'DSTYP1')) then	! DSS attributes should have been removed.
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has DSS attributes (DSTYP1), which it should not.')
!     end if
! 
!     crpixl(1) = int16Attribute( detArray, 'CRPIX1L')
!     crpixl(2) = int16Attribute( detArray, 'CRPIX2L')
!     crvall(1) = real64Attribute(detArray, 'CRVAL1L')
!     crvall(2) = real64Attribute(detArray, 'CRVAL2L')
!     cdeltl(1) = real64Attribute(detArray, 'CDELT1L')
!     cdeltl(2) = real64Attribute(detArray, 'CDELT2L')
!     if (crpixl(1) * crpixl(2) .ne. 1) then
!       print *,'CRPIXL=', crpixl
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong CRPIXL.')
!     else if ((nint(crvall(1)) .ne. -20440) .or. &	! calc '-512*80/2+80/2' - the latter is for half a pixel for the corresponding physical coordinates at CRPIXL.
!              (nint(crvall(2)) .ne. -20440)) then
!       print *,'CRVALL=', crvall
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong CRVALL.')
!     else if ((nint(cdeltl(1)) .ne. 80) .or. &
!              (nint(cdeltl(2)) .ne. 80)) then
!       print *,'CDELTL=', cdeltl
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong CDELTL.')
!     end if
! 
!     ! Tests for the hearder attributes inherited from the reference FITS file.
!     if (.not. hasAttribute(detSet, 'OBS_ID') .or. &
!         .not. hasAttribute(detSet, 'INSTRUME')) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has no OBS_ID or INSTRUME attribute.')
!     end if
! 
!     outObs_id = ''
!     outInstrument = ''
!     outObs_id     = stringAttribute( detSet, 'OBS_ID')
!     outInstrument = stringAttribute( detSet, 'INSTRUME')
! 
!     if (trim(outObs_id) .ne. trim(refObs_id)) then
!       print *,'outObs_id=', trim(outObs_id)
!       print *,'refObs_id=', trim(refObs_id)
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong OBS_ID.')
!     else if (trim(outInstrument) .ne. trim(refInstrument)) then
!       print *,'outInstrument=', trim(outInstrument)
!       print *,'refInstrument=', trim(refInstrument)
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong INSTRUME.')
!     end if
! 
!     ! Tests for the WCS hearder attributes (copied from the reference FITS file).
!     crval(1)  = real64Attribute(detArray, 'CRVAL1')
!     crval(2)  = real64Attribute(detArray, 'CRVAL2')
!     if (.not. isNearlyEqual(crval(1), refWcs%x%refWorld, 6) .or. &	! defined in test_utils
!         .not. isNearlyEqual(crval(2), refWcs%y%refWorld, 6)) then
!       print *,'CRVAL(refWcs)=', refWcs%x%refWorld, refWcs%y%refWorld
!       print *,'CRVAL=', crval
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong WCS.')
!     end if
! 
!     ! Tests for the image array itself.
!     call readArrayData(detSet, arySglRead)
!     if ( any(arySglRead(256,:) .ne. 1.0) .or. &
!          any(arySglRead(1:255,:) .ne. 0.0) ) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has wrong values.')
!     end if
!     deallocate(arySglRead)
!     call release(detSet)
!     deallocate(arySgl)
! 
! 
!     !----------------------------------------------------------
!     errHead = "SaveDetImage005a"
!     print *,"=== "//trim(errHead)//": With wcsExtended and strInstrum (single)."
! 
!     allocate(arySgl(512,512))
! 
!     arySgl(:,:)   = 0.0
!     arySgl(256,:) = 1.0
!     outFilename = 'outfiles/detSgl_im005a.ds'
!     ! outFilename = 'outfiles/detSgl_im.ds'
!     refFilename = 'outfiles/pn_2000.ds'	! Should have been prepared by another executable in this test harness!
! 
!     refSet = dataSet(refFilename, READ)
!     refObs_id = ''
!     refInstrument = ''
!     refObs_id     = stringAttribute( refSet, 'OBS_ID')
!     refInstrument = stringAttribute( refSet, 'INSTRUME')
! ! print *,'before readW'
! !     call readWcs(refSet, refWcs)	! There is a bug in ssclib/wcs_aux.f90/readWcs (for the interface to read DataSet) !!!!!
! ! print *,'after readW'
!     call release(refSet)
!     call readWcs(refFilename, refWcs)	! Move above, after readWcs is fixed.
! 
!     call convertWcsToExtended(refWcs, wcsExtended)	! defined in wcs_aux
!     wcsExtended%CRPIXL = (/ 77, 77 /)
!     wcsExtended%CRVALL(2) = -777.77d0
!     wcsExtended%withPhysical = .true.
! 
! !call dumpWcsExtended(wcsExtended, verbosityArg=0_int32, varName='wcsExtended_005')
! 
!     !call writeWcsExtendedArray(arySgl, trim(outFilename), templateFitsSetName=trim(refFilename) &
!     !   , detWcs=refWcs, strInstrum='TESTINST')
! !print *,'DEBUG: image saving.'
!     call writeWcsExtendedArray(arySgl, trim(outFilename), wcsExtended=wcsExtended &
!        , strInstrum='TESTINST')
! !print *,'DEBUG: image saved.'
! 
!     ! Test whether the output file exists or not.
!     inquire(file=trim(outFilename), exist=flagExist)
!     if ( .not. flagExist) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') not exists.')
!     end if
! 
!     detSet = dataSet(outFilename, READ)
!     detArray = array(detSet, 0, READ)
! 
!     ! Tests for the hearder attributes.
!     if (.not. hasAttribute(detArray, 'CRPIX1L') .or. &
!         .not. hasAttribute(detArray, 'CRPIX2L') .or. &
!         .not. hasAttribute(detArray, 'CRVAL1L')) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') do not have required header attributes.')
!     end if
! 
!     if (hasAttribute(detArray, 'DSTYP1')) then	! DSS attributes should not exist.
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has DSS attributes (DSTYP1), which it should not.')
!     end if
! 
!     crpixl(1) = int16Attribute( detArray, 'CRPIX1L')
!     crpixl(2) = int16Attribute( detArray, 'CRPIX2L')
!     if (int(crpixl(1), int32) * crpixl(2) .ne. 77_int32*77) then
!       print *,'CRPIXL=', crpixl
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong CRPIXL.')
!     end if
! 
!     ! Tests for the WCS hearder attributes (copied from the reference FITS file).
!     crval(1)  = real64Attribute(detArray, 'CRVAL1')
!     crval(2)  = real64Attribute(detArray, 'CRVAL2')
!     if (.not. isNearlyEqual(crval(1), refWcs%x%refWorld, 6) .or. &	! defined in test_utils
!         .not. isNearlyEqual(crval(2), refWcs%y%refWorld, 6)) then
!       print *,'CRVAL(refWcs)=', refWcs%x%refWorld, refWcs%y%refWorld
!       print *,'CRVAL=', crval
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong WCS.')
!     end if
! 
!     ! Tests for the hearder attributes of TELESCOP and INSTRUME.
!     if (.not. hasAttribute(detSet, 'TELESCOP') .or. &
!         .not. hasAttribute(detSet, 'INSTRUME')) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has no TELESCOP or INSTRUME attribute.')
!     end if
! 
!     outTelescop = ''
!     outInstrument = ''
!     outTelescop   = stringAttribute( detSet, 'TELESCOP')
!     outInstrument = stringAttribute( detSet, 'INSTRUME')
! 
!     if (trim(outTelescop) .ne. 'XMM') then
!       print *,'outTelescop=', trim(outTelescop)
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong TELESCOP.')
!     else if (trim(outInstrument) .ne. 'TESTINST') then
!       print *,'outInstrument=', trim(outInstrument)
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong INSTRUME.')
!     end if
! 
!     ! Tests for the image array itself.
!     call readArrayData(detSet, arySglRead)
!     if ( any(arySglRead(256,:) .ne. 1.0) .or. &
!          any(arySglRead(1:255,:) .ne. 0.0) ) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has wrong values.')
!     end if
!     deallocate(arySglRead)
!     call release(detSet)
!     deallocate(arySgl)
! 
! 
!     !----------------------------------------------------------
!     errHead = "SaveDetImage105a"
!     print *,"=== "//trim(errHead)//": With wcsExtended and strInstrum (double)."
! 
!     allocate(aryDbl(512,512))
! 
!     aryDbl(:,:)   = 0.0
!     aryDbl(256,:) = 1.0
!     outFilename = 'outfiles/detDbl_im.ds'
!     refFilename = 'outfiles/pn_2000.ds'	! Should have been prepared by another executable in this test harness!
! 
!     refSet = dataSet(refFilename, READ)
!     refObs_id = ''
!     refInstrument = ''
!     refObs_id     = stringAttribute( refSet, 'OBS_ID')
!     refInstrument = stringAttribute( refSet, 'INSTRUME')
! ! print *,'before readW'
! !     call readWcs(refSet, refWcs)	! There is a bug in ssclib/wcs_aux.f90/readWcs (for the interface to read DataSet) !!!!!
! ! print *,'after readW'
!     call release(refSet)
!     call readWcs(refFilename, refWcs)	! Move above, after readWcs is fixed.
! 
!     call convertWcsToExtended(refWcs, wcsExtended)	! defined in wcs_aux
! 
!     call writeWcsExtendedArray(aryDbl, trim(outFilename), templateFitsSetName=trim(refFilename) &
!        , detWcs=refWcs, strInstrum='TESTINST')
! 
!     ! Test whether the output file exists or not.
!     inquire(file=trim(outFilename), exist=flagExist)
!     if ( .not. flagExist) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') not exists.')
!     end if
! 
!     detSet = dataSet(outFilename, READ)
!     detArray = array(detSet, 0, READ)
! 
!     ! Tests for the hearder attributes.
!     if (.not. hasAttribute(detArray, 'CRPIX1L') .or. &
!         .not. hasAttribute(detArray, 'CRVAL1L') .or. &
!         .not. hasAttribute(detArray, 'CDELT1L') .or. &
!         .not. hasAttribute(detArray, 'CTYPE1L') .or. &
!         .not. hasAttribute(detArray, 'CRPIX2L') .or. &
!         .not. hasAttribute(detArray, 'CRVAL2L') .or. &
!         .not. hasAttribute(detArray, 'CDELT2L') .or. &
!         .not. hasAttribute(detArray, 'CTYPE2L')) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') do not have required header attributes.')
!     end if
! 
!     if (hasAttribute(detArray, 'DSTYP1')) then	! DSS attributes should have been removed.
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has DSS attributes (DSTYP1), which it should not.')
!     end if
! 
!     crpixl(1) = int16Attribute( detArray, 'CRPIX1L')
!     crpixl(2) = int16Attribute( detArray, 'CRPIX2L')
!     crvall(1) = real64Attribute(detArray, 'CRVAL1L')
!     crvall(2) = real64Attribute(detArray, 'CRVAL2L')
!     cdeltl(1) = real64Attribute(detArray, 'CDELT1L')
!     cdeltl(2) = real64Attribute(detArray, 'CDELT2L')
!     if (crpixl(1) * crpixl(2) .ne. 1) then
!       print *,'CRPIXL=', crpixl
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong CRPIXL.')
!     else if ((nint(crvall(1)) .ne. -20440) .or. &	! calc '-512*80/2+80/2' - the latter is for half a pixel for the corresponding physical coordinates at CRPIXL.
!              (nint(crvall(2)) .ne. -20440)) then
!       print *,'CRVALL=', crvall
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong CRVALL.')
!     else if ((nint(cdeltl(1)) .ne. 80) .or. &
!              (nint(cdeltl(2)) .ne. 80)) then
!       print *,'CDELTL=', cdeltl
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong CDELTL.')
!     end if
! 
!     ! Tests for the hearder attributes inherited from the reference FITS file.
!     if (.not. hasAttribute(detSet, 'OBS_ID')) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has no OBS_ID attribute.')
!     end if
! 
!     outObs_id = ''
!     ! outInstrument = ''
!     outObs_id     = stringAttribute( detSet, 'OBS_ID')
!     ! outInstrument = stringAttribute( detSet, 'INSTRUME')
! 
!     if (trim(outObs_id) .ne. trim(refObs_id)) then
!       print *,'outObs_id=', trim(outObs_id)
!       print *,'refObs_id=', trim(refObs_id)
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong OBS_ID.')
!     end if
! 
!     ! Tests for the WCS hearder attributes (copied from the reference FITS file).
!     crval(1)  = real64Attribute(detArray, 'CRVAL1')
!     crval(2)  = real64Attribute(detArray, 'CRVAL2')
!     if (.not. isNearlyEqual(crval(1), refWcs%x%refWorld, 6) .or. &	! defined in test_utils
!         .not. isNearlyEqual(crval(2), refWcs%y%refWorld, 6)) then
!       print *,'CRVAL(refWcs)=', refWcs%x%refWorld, refWcs%y%refWorld
!       print *,'CRVAL=', crval
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong WCS.')
!     end if
! 
!     ! Tests for the hearder attributes of TELESCOP and INSTRUME.
!     if (.not. hasAttribute(detSet, 'TELESCOP') .or. &
!         .not. hasAttribute(detSet, 'INSTRUME')) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has no TELESCOP or INSTRUME attribute.')
!     end if
! 
!     outTelescop = ''
!     outInstrument = ''
!     outTelescop   = stringAttribute( detSet, 'TELESCOP')
!     outInstrument = stringAttribute( detSet, 'INSTRUME')
! 
!     if (trim(outTelescop) .ne. 'XMM') then
!       print *,'outTelescop=', trim(outTelescop)
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong TELESCOP.')
!     else if (trim(outInstrument) .ne. 'TESTINST') then
!       print *,'outInstrument=', trim(outInstrument)
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong INSTRUME.')
!     end if
! 
!     ! Tests for the image array itself.
!     call readArrayData(detSet, aryDblRead)
!     if ( any(aryDblRead(256,:) .ne. 1.0) .or. &
!          any(aryDblRead(1:255,:) .ne. 0.0) ) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has wrong values.')
!     end if
!     deallocate(aryDblRead)
!     call release(detSet)
!     deallocate(aryDbl)
! 
! 
!     !----------------------------------------------------------
!     errHead = "SaveDetImage205a"
!     print *,"=== "//trim(errHead)//": With wcsExtended and strInstrum (Int32)."
! 
!     allocate(aryInt32(512,512))
! 
!     aryInt32(:,:)   = 0.0
!     aryInt32(256,:) = 1.0
!     outFilename = 'outfiles/detInt32_im.ds'
!     refFilename = 'outfiles/pn_2000.ds'	! Should have been prepared by another executable in this test harness!
! 
!     refSet = dataSet(refFilename, READ)
!     refObs_id = ''
!     refInstrument = ''
!     refObs_id     = stringAttribute( refSet, 'OBS_ID')
!     refInstrument = stringAttribute( refSet, 'INSTRUME')
! ! print *,'before readW'
! !     call readWcs(refSet, refWcs)	! There is a bug in ssclib/wcs_aux.f90/readWcs (for the interface to read DataSet) !!!!!
! ! print *,'after readW'
!     call release(refSet)
!     call readWcs(refFilename, refWcs)	! Move above, after readWcs is fixed.
! 
!     call convertWcsToExtended(refWcs, wcsExtended)	! defined in wcs_aux
! 
!     call writeWcsExtendedArray(aryInt32, trim(outFilename), templateFitsSetName=trim(refFilename) &
!        , detWcs=refWcs, strInstrum='TESTINST')
! 
!     ! Test whether the output file exists or not.
!     inquire(file=trim(outFilename), exist=flagExist)
!     if ( .not. flagExist) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') not exists.')
!     end if
! 
!     detSet = dataSet(outFilename, READ)
!     detArray = array(detSet, 0, READ)
! 
!     ! Tests for the hearder attributes.
!     if (.not. hasAttribute(detArray, 'CRPIX1L') .or. &
!         .not. hasAttribute(detArray, 'CRVAL1L') .or. &
!         .not. hasAttribute(detArray, 'CDELT1L') .or. &
!         .not. hasAttribute(detArray, 'CTYPE1L') .or. &
!         .not. hasAttribute(detArray, 'CRPIX2L') .or. &
!         .not. hasAttribute(detArray, 'CRVAL2L') .or. &
!         .not. hasAttribute(detArray, 'CDELT2L') .or. &
!         .not. hasAttribute(detArray, 'CTYPE2L')) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') do not have required header attributes.')
!     end if
! 
!     if (hasAttribute(detArray, 'DSTYP1')) then	! DSS attributes should have been removed.
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has DSS attributes (DSTYP1), which it should not.')
!     end if
! 
!     crpixl(1) = int16Attribute( detArray, 'CRPIX1L')
!     crpixl(2) = int16Attribute( detArray, 'CRPIX2L')
!     crvall(1) = real64Attribute(detArray, 'CRVAL1L')
!     crvall(2) = real64Attribute(detArray, 'CRVAL2L')
!     cdeltl(1) = real64Attribute(detArray, 'CDELT1L')
!     cdeltl(2) = real64Attribute(detArray, 'CDELT2L')
!     if (crpixl(1) * crpixl(2) .ne. 1) then
!       print *,'CRPIXL=', crpixl
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong CRPIXL.')
!     else if ((nint(crvall(1)) .ne. -20440) .or. &	! calc '-512*80/2+80/2' - the latter is for half a pixel for the corresponding physical coordinates at CRPIXL.
!              (nint(crvall(2)) .ne. -20440)) then
!       print *,'CRVALL=', crvall
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong CRVALL.')
!     else if ((nint(cdeltl(1)) .ne. 80) .or. &
!              (nint(cdeltl(2)) .ne. 80)) then
!       print *,'CDELTL=', cdeltl
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong CDELTL.')
!     end if
! 
!     ! Tests for the hearder attributes inherited from the reference FITS file.
!     if (.not. hasAttribute(detSet, 'OBS_ID')) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has no OBS_ID attribute.')
!     end if
! 
!     outObs_id = ''
!     ! outInstrument = ''
!     outObs_id     = stringAttribute( detSet, 'OBS_ID')
!     ! outInstrument = stringAttribute( detSet, 'INSTRUME')
! 
!     if (trim(outObs_id) .ne. trim(refObs_id)) then
!       print *,'outObs_id=', trim(outObs_id)
!       print *,'refObs_id=', trim(refObs_id)
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong OBS_ID.')
!     end if
! 
!     ! Tests for the WCS hearder attributes (copied from the reference FITS file).
!     crval(1)  = real64Attribute(detArray, 'CRVAL1')
!     crval(2)  = real64Attribute(detArray, 'CRVAL2')
!     if (.not. isNearlyEqual(crval(1), refWcs%x%refWorld, 6) .or. &	! defined in test_utils
!         .not. isNearlyEqual(crval(2), refWcs%y%refWorld, 6)) then
!       print *,'CRVAL(refWcs)=', refWcs%x%refWorld, refWcs%y%refWorld
!       print *,'CRVAL=', crval
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong WCS.')
!     end if
! 
!     ! Tests for the hearder attributes of TELESCOP and INSTRUME.
!     if (.not. hasAttribute(detSet, 'TELESCOP') .or. &
!         .not. hasAttribute(detSet, 'INSTRUME')) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has no TELESCOP or INSTRUME attribute.')
!     end if
! 
!     outTelescop = ''
!     outInstrument = ''
!     outTelescop   = stringAttribute( detSet, 'TELESCOP')
!     outInstrument = stringAttribute( detSet, 'INSTRUME')
! 
!     if (trim(outTelescop) .ne. 'XMM') then
!       print *,'outTelescop=', trim(outTelescop)
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong TELESCOP.')
!     else if (trim(outInstrument) .ne. 'TESTINST') then
!       print *,'outInstrument=', trim(outInstrument)
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong INSTRUME.')
!     end if
! 
!     ! Tests for the image array itself.
!     call readArrayData(detSet, aryInt32Read)
!     if ( any(aryInt32Read(256,:) .ne. 1.0) .or. &
!          any(aryInt32Read(1:255,:) .ne. 0.0) ) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has wrong values.')
!     end if
!     deallocate(aryInt32Read)
!     call release(detSet)
!     deallocate(aryInt32)
! 
! 
!     !----------------------------------------------------------
!     errHead = "SaveDetImage305a"
!     print *,"=== "//trim(errHead)//": With wcsExtended and strInstrum (Int16)."
! 
!     allocate(aryInt16(512,512))
! 
!     aryInt16(:,:)   = 0.0
!     aryInt16(256,:) = 1.0
!     outFilename = 'outfiles/detInt16_im.ds'
!     refFilename = 'outfiles/pn_2000.ds'	! Should have been prepared by another executable in this test harness!
! 
!     refSet = dataSet(refFilename, READ)
!     refObs_id = ''
!     refInstrument = ''
!     refObs_id     = stringAttribute( refSet, 'OBS_ID')
!     refInstrument = stringAttribute( refSet, 'INSTRUME')
! ! print *,'before readW'
! !     call readWcs(refSet, refWcs)	! There is a bug in ssclib/wcs_aux.f90/readWcs (for the interface to read DataSet) !!!!!
! ! print *,'after readW'
!     call release(refSet)
!     call readWcs(refFilename, refWcs)	! Move above, after readWcs is fixed.
! 
!     call convertWcsToExtended(refWcs, wcsExtended)	! defined in wcs_aux
! 
!     call writeWcsExtendedArray(aryInt16, trim(outFilename), templateFitsSetName=trim(refFilename) &
!        , detWcs=refWcs, strInstrum='TESTINST')
! 
!     ! Test whether the output file exists or not.
!     inquire(file=trim(outFilename), exist=flagExist)
!     if ( .not. flagExist) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') not exists.')
!     end if
! 
!     detSet = dataSet(outFilename, READ)
!     detArray = array(detSet, 0, READ)
! 
!     ! Tests for the hearder attributes.
!     if (.not. hasAttribute(detArray, 'CRPIX1L') .or. &
!         .not. hasAttribute(detArray, 'CRVAL1L') .or. &
!         .not. hasAttribute(detArray, 'CDELT1L') .or. &
!         .not. hasAttribute(detArray, 'CTYPE1L') .or. &
!         .not. hasAttribute(detArray, 'CRPIX2L') .or. &
!         .not. hasAttribute(detArray, 'CRVAL2L') .or. &
!         .not. hasAttribute(detArray, 'CDELT2L') .or. &
!         .not. hasAttribute(detArray, 'CTYPE2L')) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') do not have required header attributes.')
!     end if
! 
!     if (hasAttribute(detArray, 'DSTYP1')) then	! DSS attributes should have been removed.
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has DSS attributes (DSTYP1), which it should not.')
!     end if
! 
!     crpixl(1) = int16Attribute( detArray, 'CRPIX1L')
!     crpixl(2) = int16Attribute( detArray, 'CRPIX2L')
!     crvall(1) = real64Attribute(detArray, 'CRVAL1L')
!     crvall(2) = real64Attribute(detArray, 'CRVAL2L')
!     cdeltl(1) = real64Attribute(detArray, 'CDELT1L')
!     cdeltl(2) = real64Attribute(detArray, 'CDELT2L')
!     if (crpixl(1) * crpixl(2) .ne. 1) then
!       print *,'CRPIXL=', crpixl
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong CRPIXL.')
!     else if ((nint(crvall(1)) .ne. -20440) .or. &	! calc '-512*80/2+80/2' - the latter is for half a pixel for the corresponding physical coordinates at CRPIXL.
!              (nint(crvall(2)) .ne. -20440)) then
!       print *,'CRVALL=', crvall
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong CRVALL.')
!     else if ((nint(cdeltl(1)) .ne. 80) .or. &
!              (nint(cdeltl(2)) .ne. 80)) then
!       print *,'CDELTL=', cdeltl
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong CDELTL.')
!     end if
! 
!     ! Tests for the hearder attributes inherited from the reference FITS file.
!     if (.not. hasAttribute(detSet, 'OBS_ID')) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has no OBS_ID attribute.')
!     end if
! 
!     outObs_id = ''
!     ! outInstrument = ''
!     outObs_id     = stringAttribute( detSet, 'OBS_ID')
!     ! outInstrument = stringAttribute( detSet, 'INSTRUME')
! 
!     if (trim(outObs_id) .ne. trim(refObs_id)) then
!       print *,'outObs_id=', trim(outObs_id)
!       print *,'refObs_id=', trim(refObs_id)
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong OBS_ID.')
!     end if
! 
!     ! Tests for the WCS hearder attributes (copied from the reference FITS file).
!     crval(1)  = real64Attribute(detArray, 'CRVAL1')
!     crval(2)  = real64Attribute(detArray, 'CRVAL2')
!     if (.not. isNearlyEqual(crval(1), refWcs%x%refWorld, 6) .or. &	! defined in test_utils
!         .not. isNearlyEqual(crval(2), refWcs%y%refWorld, 6)) then
!       print *,'CRVAL(refWcs)=', refWcs%x%refWorld, refWcs%y%refWorld
!       print *,'CRVAL=', crval
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong WCS.')
!     end if
! 
!     ! Tests for the hearder attributes of TELESCOP and INSTRUME.
!     if (.not. hasAttribute(detSet, 'TELESCOP') .or. &
!         .not. hasAttribute(detSet, 'INSTRUME')) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has no TELESCOP or INSTRUME attribute.')
!     end if
! 
!     outTelescop = ''
!     outInstrument = ''
!     outTelescop   = stringAttribute( detSet, 'TELESCOP')
!     outInstrument = stringAttribute( detSet, 'INSTRUME')
! 
!     if (trim(outTelescop) .ne. 'XMM') then
!       print *,'outTelescop=', trim(outTelescop)
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong TELESCOP.')
!     else if (trim(outInstrument) .ne. 'TESTINST') then
!       print *,'outInstrument=', trim(outInstrument)
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong INSTRUME.')
!     end if
! 
!     ! Tests for the image array itself.
!     call readArrayData(detSet, aryInt16Read)
!     if ( any(aryInt16Read(256,:) .ne. 1.0) .or. &
!          any(aryInt16Read(1:255,:) .ne. 0.0) ) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has wrong values.')
!     end if
!     deallocate(aryInt16Read)
!     call release(detSet)
!     deallocate(aryInt16)
! 
! 
!     !----------------------------------------------------------
!     errHead = "SaveDetImage405a"
!     print *,"=== "//trim(errHead)//": With wcsExtended and strInstrum (Int8)."
! 
!     allocate(aryInt8(512,512))
! 
!     aryInt8(:,:)   = 0.0
!     aryInt8(256,:) = 1.0
!     outFilename = 'outfiles/detInt8_im.ds'
!     refFilename = 'outfiles/pn_2000.ds'	! Should have been prepared by another executable in this test harness!
! 
!     refSet = dataSet(refFilename, READ)
!     refObs_id = ''
!     refInstrument = ''
!     refObs_id     = stringAttribute( refSet, 'OBS_ID')
!     refInstrument = stringAttribute( refSet, 'INSTRUME')
! ! print *,'before readW'
! !     call readWcs(refSet, refWcs)	! There is a bug in ssclib/wcs_aux.f90/readWcs (for the interface to read DataSet) !!!!!
! ! print *,'after readW'
!     call release(refSet)
!     call readWcs(refFilename, refWcs)	! Move above, after readWcs is fixed.
! 
!     call convertWcsToExtended(refWcs, wcsExtended)	! defined in wcs_aux
! 
!     call writeWcsExtendedArray(aryInt8, trim(outFilename), templateFitsSetName=trim(refFilename) &
!        , detWcs=refWcs, strInstrum='TESTINST')
! 
!     ! Test whether the output file exists or not.
!     inquire(file=trim(outFilename), exist=flagExist)
!     if ( .not. flagExist) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') not exists.')
!     end if
! 
!     detSet = dataSet(outFilename, READ)
!     detArray = array(detSet, 0, READ)
! 
!     ! Tests for the hearder attributes.
!     if (.not. hasAttribute(detArray, 'CRPIX1L') .or. &
!         .not. hasAttribute(detArray, 'CRVAL1L') .or. &
!         .not. hasAttribute(detArray, 'CDELT1L') .or. &
!         .not. hasAttribute(detArray, 'CTYPE1L') .or. &
!         .not. hasAttribute(detArray, 'CRPIX2L') .or. &
!         .not. hasAttribute(detArray, 'CRVAL2L') .or. &
!         .not. hasAttribute(detArray, 'CDELT2L') .or. &
!         .not. hasAttribute(detArray, 'CTYPE2L')) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') do not have required header attributes.')
!     end if
! 
!     if (hasAttribute(detArray, 'DSTYP1')) then	! DSS attributes should have been removed.
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has DSS attributes (DSTYP1), which it should not.')
!     end if
! 
!     crpixl(1) = int16Attribute( detArray, 'CRPIX1L')
!     crpixl(2) = int16Attribute( detArray, 'CRPIX2L')
!     crvall(1) = real64Attribute(detArray, 'CRVAL1L')
!     crvall(2) = real64Attribute(detArray, 'CRVAL2L')
!     cdeltl(1) = real64Attribute(detArray, 'CDELT1L')
!     cdeltl(2) = real64Attribute(detArray, 'CDELT2L')
!     if (crpixl(1) * crpixl(2) .ne. 1) then
!       print *,'CRPIXL=', crpixl
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong CRPIXL.')
!     else if ((nint(crvall(1)) .ne. -20440) .or. &	! calc '-512*80/2+80/2' - the latter is for half a pixel for the corresponding physical coordinates at CRPIXL.
!              (nint(crvall(2)) .ne. -20440)) then
!       print *,'CRVALL=', crvall
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong CRVALL.')
!     else if ((nint(cdeltl(1)) .ne. 80) .or. &
!              (nint(cdeltl(2)) .ne. 80)) then
!       print *,'CDELTL=', cdeltl
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong CDELTL.')
!     end if
! 
!     ! Tests for the hearder attributes inherited from the reference FITS file.
!     if (.not. hasAttribute(detSet, 'OBS_ID')) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has no OBS_ID attribute.')
!     end if
! 
!     outObs_id = ''
!     ! outInstrument = ''
!     outObs_id     = stringAttribute( detSet, 'OBS_ID')
!     ! outInstrument = stringAttribute( detSet, 'INSTRUME')
! 
!     if (trim(outObs_id) .ne. trim(refObs_id)) then
!       print *,'outObs_id=', trim(outObs_id)
!       print *,'refObs_id=', trim(refObs_id)
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong OBS_ID.')
!     end if
! 
!     ! Tests for the WCS hearder attributes (copied from the reference FITS file).
!     crval(1)  = real64Attribute(detArray, 'CRVAL1')
!     crval(2)  = real64Attribute(detArray, 'CRVAL2')
!     if (.not. isNearlyEqual(crval(1), refWcs%x%refWorld, 6) .or. &	! defined in test_utils
!         .not. isNearlyEqual(crval(2), refWcs%y%refWorld, 6)) then
!       print *,'CRVAL(refWcs)=', refWcs%x%refWorld, refWcs%y%refWorld
!       print *,'CRVAL=', crval
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong WCS.')
!     end if
! 
!     ! Tests for the hearder attributes of TELESCOP and INSTRUME.
!     if (.not. hasAttribute(detSet, 'TELESCOP') .or. &
!         .not. hasAttribute(detSet, 'INSTRUME')) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has no TELESCOP or INSTRUME attribute.')
!     end if
! 
!     outTelescop = ''
!     outInstrument = ''
!     outTelescop   = stringAttribute( detSet, 'TELESCOP')
!     outInstrument = stringAttribute( detSet, 'INSTRUME')
! 
!     if (trim(outTelescop) .ne. 'XMM') then
!       print *,'outTelescop=', trim(outTelescop)
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong TELESCOP.')
!     else if (trim(outInstrument) .ne. 'TESTINST') then
!       print *,'outInstrument=', trim(outInstrument)
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has a wrong INSTRUME.')
!     end if
! 
!     ! Tests for the image array itself.
!     call readArrayData(detSet, aryInt8Read)
!     if ( any(aryInt8Read(256,:) .ne. 1.0) .or. &
!          any(aryInt8Read(1:255,:) .ne. 0.0) ) then
!       call error("err"//trim(errHead), 'File('//trim(outFilename)//') has wrong values.')
!     end if
!     deallocate(aryInt8Read)
!     call release(detSet)
!     deallocate(aryInt8)
! 
! 
! 
  end subroutine test_writeWcsExtendedArray

end program testwcsaux
